"""File pareto.py

:author: Michel Bierlaire, EPFL
:date: Tue Jul  4 17:11:29 2023

Implement a Pareto set for generic purposes.
"""

from __future__ import annotations
import logging
from datetime import datetime
import tomlkit as tk
from matplotlib.axes import Axes
from tomlkit.exceptions import NonExistentKey

from biogeme_optimization.exceptions import OptimizationError

try:
    import matplotlib.pyplot as plt

    CAN_PLOT = True
except ModuleNotFoundError:
    CAN_PLOT = False

DATE_TIME_STRING = '__DATETIME__'


def replace_date_time(a_string: str) -> str:
    """Replaces the string defined above by the current time and date

    :param a_string: the string to be modified

    :return: the modified string, if the modification has been made. None otherwise
    """
    current_datetime = datetime.now()
    formatted_datetime = current_datetime.strftime("%B %d, %Y. %H:%M:%S")
    return a_string.replace(f'{DATE_TIME_STRING}', formatted_datetime)


logger = logging.getLogger(__name__)


class SetElement:
    """Specify the elements of the Pareto set. Note that each
    objective is supposed to be minimized.

    """

    def __init__(self, element_id: str, objectives: list[float]) -> None:
        """Ctor

        :param element_id: identifier of the element
        :type element_id: str

        :param objectives: values of the objective functions
        :type objectives: list(float)
        """
        self.element_id: str = element_id
        self.objectives: list[float] = objectives

        if any(obj is None for obj in objectives):
            raise OptimizationError(f'All objectives must be defined: {objectives}')

    def __eq__(self, other: SetElement) -> bool:
        if isinstance(other, SetElement):
            if self.element_id == other.element_id:
                if self.objectives != other.objectives:
                    error_msg = (
                        f'Two elements named {self.element_id} have different '
                        f'objective values: {self.objectives} and '
                        f'{other.objectives}'
                    )
                    raise OptimizationError(error_msg)
                return True
        return False

    def __hash__(self) -> int:
        return hash((self.element_id, tuple(self.objectives)))

    def __str__(self) -> str:
        return f'{self.element_id} {self.objectives}'

    def __repr__(self) -> str:
        return self.element_id

    def dominates(self, other: SetElement) -> bool:
        """Verifies if self dominates other.

        :param other: other element to check
        :type other: SetElement
        """
        if len(self.objectives) != len(other.objectives):
            raise OptimizationError(
                f'Incompatible sizes: '
                f'{len(self.objectives)}'
                f' and {len(other.objectives)}'
            )

        if any(my_f > her_f for my_f, her_f in zip(self.objectives, other.objectives)):
            return False

        return self.objectives != other.objectives


class Pareto:
    """This object manages a Pareto set for a list of objectives that
    are each minimized.
    """

    def __init__(self, filename: str | None = None) -> None:
        self.size_init_pareto: int = 0
        self.size_init_considered: int = 0
        self.size_init_invalid: int = 0
        self.filename: str | None = filename
        self.comments: list[str] = [f'File automatically created on {DATE_TIME_STRING}']
        """Comment to be inserted in the file when dumped, where
            __DATETIME__ is replaced by the current date and time
        
        """
        self.pareto: set[SetElement] = set()
        """set of elements
        """

        self.removed: set[SetElement] = set()
        """set of solutions that have been in the Pareto set ar some point,
            but have been removed because dominated by another
            solution.
        """

        self.considered: set[SetElement] = set()
        """set of solutions that have been considered at some point by the
            algorithm
        """

        self.invalid: set[SetElement] = set()
        """set of solutions that have been deemed invalid
        """

        if filename is not None:
            if self.restore():
                logger.debug('RESTORE PARETO FROM FILE')
                logger.info(
                    f'Pareto set initialized from file with '
                    f'{self.size_init_considered} elements '
                    f'[{self.size_init_pareto} Pareto] and '
                    f'{self.size_init_invalid} invalid elements.'
                )
                logger.debug('RESTORE PARETO FROM FILE: DONE')

            else:
                logger.info(f'Unable to read file {filename}. Pareto set empty.')

    def __str__(self) -> str:
        return (
            f'Pareto: {self.pareto} Removed: {self.removed} '
            f'Considered: {self.considered} Invalid: {self.invalid}'
        )

    def dump(self) -> None:
        """
        Dump the three sets on a file

        :raise OptimizationError: if a problem has occured during dumping.
        """
        if self.filename is None:
            logger.warning('No Pareto file has been provided')
            return
        doc = tk.document()
        final_comments = [replace_date_time(comment) for comment in self.comments]
        for comment in final_comments:
            doc.add(tk.comment(comment))

        pareto_table = tk.table()
        for elem in self.pareto:
            pareto_table[elem.element_id] = [float(obj) for obj in elem.objectives]
        doc['Pareto'] = pareto_table

        considered_table = tk.table()
        for elem in self.considered:
            considered_table[elem.element_id] = [float(obj) for obj in elem.objectives]
        doc['Considered'] = considered_table

        removed_table = tk.table()
        for elem in self.removed:
            removed_table[elem.element_id] = [float(obj) for obj in elem.objectives]
        doc['Removed'] = removed_table

        invalid_table = tk.table()
        for elem in self.invalid:
            invalid_table[elem.element_id] = [float(obj) for obj in elem.objectives]
        doc['Invalid'] = invalid_table

        with open(self.filename, 'w', encoding='utf-8') as f:
            print(tk.dumps(doc), file=f)

    def get_element_from_id(self, the_id: str) -> SetElement | None:
        """Returns the element of a set given its ID

        :param the_id: identifiers of the element to return
        :type the_id: str

        :return: found element, or None if element not present
        :rtype: SetElement
        """
        found = {elem for elem in self.considered if elem.element_id == the_id}
        if len(found) == 0:
            return None
        if len(found) > 1:
            error_msg = f'There are {len(found)} elements with ID {the_id}'
            raise OptimizationError(error_msg)
        return next(iter(found))

    def parse_set_from_document(
        self, document: tk.document, set_name: str
    ) -> set[SetElement]:
        """Parse a set from the document based on the set name.

        :param document: parsed document
        :type document: tomlkit.document

        :param set_name: name of the set to extract
        :type set_name: str

        :return: set of SetElement objects
        :rtype: set
        """
        try:
            the_set = document[set_name]
        except NonExistentKey:
            logger.warning(f'No {set_name} section in pareto file')
            return set()
        return {SetElement(the_id, list(values)) for the_id, values in the_set.items()}

    def restore(self) -> bool:
        """Restore the Pareto from a file

        :return: True if the set was properly restored. False if the file was not found.
        """
        try:
            with open(self.filename, 'r', encoding='utf-8') as f:
                content = f.read()
                document = tk.parse(content)
        except FileNotFoundError:
            return False

        self.pareto = self.parse_set_from_document(document, 'Pareto')
        self.considered = self.parse_set_from_document(document, 'Considered')
        self.removed = self.parse_set_from_document(document, 'Removed')
        self.invalid = self.parse_set_from_document(document, 'Invalid')

        self.size_init_pareto = len(self.pareto)
        self.size_init_considered = len(self.considered)
        self.size_init_invalid = len(self.invalid)
        return True

    def length(self) -> int:
        """
        Obtain the length of the pareto set.
        """
        return len(self.pareto)

    def length_of_all_sets(self) -> tuple[int, int, int, int]:
        """
        Obtain the length of the four sets.
        """
        return (
            len(self.pareto),
            len(self.considered),
            len(self.removed),
            len(self.invalid),
        )

    def add_invalid(self, element: SetElement) -> bool:
        """

        :param element: invalid element to be stored
        :type element: solutionClass

        :return: True if element has been included. False otherwise.
        :rtype: bool

        """
        if element in self.invalid:
            warning_msg = f'Invalid element {element.element_id} has already been inserted in the set'
            logger.debug(warning_msg)
            return False

        self.invalid.add(element)

    def add(self, element: SetElement) -> bool:
        """

        - We define the set D as the set of members of the current
          Pareto set that are dominated by the element elem.

        - We define the set S as the set of members of the current
          Pareto set that dominate the element elem.

        If S is empty, we add elem to the Pareto set, and remove all
        elements in D.

        :param element: element to be considered for inclusion in the Pareto set.
        :type element: solutionClass

        :return: True if elemenet has been included. False otherwise.
        :rtype: bool

        """
        if element in self.considered:
            warning_msg = (
                f'Elem {element.element_id} has already been inserted in the set'
            )
            logger.debug(warning_msg)
            return False

        self.considered.add(element)
        s_dominated = set()
        d_dominating = set()
        for k in self.pareto:
            if element.dominates(k):
                d_dominating.add(k)
            if k.dominates(element):
                s_dominated.add(k)
        if s_dominated:
            return False
        self.pareto.add(element)
        self.pareto = {k for k in self.pareto if k not in d_dominating}
        self.removed |= d_dominating
        return True

    def statistics(self) -> tuple[str, str, str]:
        """Report some statistics about the Pareto set

        :return: tuple of messages, possibly empty.
        :rtype: tuple(str)
        """
        if self.pareto is None:
            return '', '', ''
        msg = (
            f'Pareto: {len(self.pareto)} ',
            f'Considered: {len(self.considered)} ',
            f'Removed: {len(self.removed)}',
        )
        return msg

    def plot(
        self,
        objective_x: int = 0,
        objective_y: int = 1,
        label_x: str | None = None,
        label_y: str | None = None,
        margin_x: int = 5,
        margin_y: int = 5,
        ax: Axes | None = None,
    ):
        """Plot the members of the set according to two
            objective functions.  They  determine the x- and
            y-coordinate of the plot.

        :param objective_x: index of the objective function to use for the x-coordinate.
        :param objective_x: int

        :param objective_y: index of the objective function to use for the y-coordinate.
        :param objective_y: int

        :param label_x: label for the x-axis
        :type label_x: str

        :param label_y: label for the y-axis
        :type label_y: str

        :param margin_x: margin for the x-axis
        :type margin_x: int

        :param margin_y: margin for the y-axis
        :type margin_y: int

        :param ax: matplotlib axis for the plot
        :type ax: matplotlib.Axes

        """
        if not CAN_PLOT:
            raise OptimizationError('Install matplotlib.')
        ax = ax or plt.gca()

        if self.length() == 0:
            raise OptimizationError('Cannot plot an empty Pareto set')

        first_elem = next(iter(self.pareto))
        number_of_objectives = len(first_elem.objectives)

        if number_of_objectives < 2:
            raise OptimizationError(
                'At least two objectives functions are required for the plot of '
                'the Pateto set.'
            )

        if objective_x >= number_of_objectives:
            error_msg = (
                f'Index of objective x is {objective_x}, but there are '
                f'only {number_of_objectives}. Give a number between 0 '
                'and {number_of_objectives-1}.'
            )
            raise OptimizationError(error_msg)

        par_x = [elem.objectives[objective_x] for elem in self.pareto]
        par_y = [elem.objectives[objective_y] for elem in self.pareto]

        con_x = [elem.objectives[objective_x] for elem in self.considered]
        con_y = [elem.objectives[objective_y] for elem in self.considered]

        rem_x = [elem.objectives[objective_x] for elem in self.removed]
        rem_y = [elem.objectives[objective_y] for elem in self.removed]

        inv_x = [elem.objectives[objective_x] for elem in self.invalid]
        inv_y = [elem.objectives[objective_y] for elem in self.invalid]

        ax.axis(
            [
                min(par_x) - margin_x,
                max(par_x) + margin_x,
                min(par_y) - margin_y,
                max(par_y) + margin_y,
            ]
        )
        ax.plot(par_x, par_y, 'o', label='Pareto')
        ax.plot(rem_x, rem_y, 'x', label='Removed')
        ax.plot(con_x, con_y, ',', label='Considered')
        if self.invalid:
            ax.plot(inv_x, inv_y, '*', label='Invalid')
        if label_x is None:
            label_x = f'Objective {objective_x}'
        if label_y is None:
            label_y = f'Objective {objective_y}'

        ax.set_xlabel(label_x)
        ax.set_ylabel(label_y)
        ax.legend()
        return ax



"""File test_linesearch.py

:author: Michel Bierlaire
:date: Thu Jun 22 16:04:39 2023

Tests for the linesearch module
"""

import unittest
import numpy as np
from biogeme_optimization.exceptions import (
    OptimizationError,
)  # Import the required exception class
from biogeme_optimization.linesearch import (
    line_search,
    newton_line_search,
    bfgs_line_search,
    DirectionManagement,
    NewtonDirection,
    InverseBfgsDirection,
)
from examples import MyFunctionToMinimize, Example58, Rosenbrock


class TestDirectionManagement(unittest.TestCase):
    def test_abstract_class(self):
        fct = MyFunctionToMinimize(dimension=2)
        with self.assertRaises(TypeError):
            _ = DirectionManagement(function=fct)


class TestNewtonDirection(unittest.TestCase):
    def test_get_direction(self):
        fct = MyFunctionToMinimize(dimension=2)
        newton_direction = NewtonDirection(function=fct)
        iterate = np.array([1, 1])
        the_dir = newton_direction.get_direction(iterate)
        expected_dir = np.array([-1, -1])
        np.testing.assert_array_almost_equal(the_dir, expected_dir, decimal=6)

        optimal_solution = np.array([0, 0])
        the_dir = newton_direction.get_direction(optimal_solution)
        self.assertIsNone(the_dir)


class TestInverseBfgsDirection(unittest.TestCase):
    def test_get_direction(self):
        fct = MyFunctionToMinimize(dimension=2)
        inverse_bfgs_direction = InverseBfgsDirection(function=fct)
        iterate = np.array([1, 1])
        the_dir = inverse_bfgs_direction.get_direction(iterate)
        expected_dir = np.array([-1, -1])
        np.testing.assert_array_almost_equal(the_dir, expected_dir, decimal=6)

        iterate = np.array([0.5, 0.5])
        the_dir = inverse_bfgs_direction.get_direction(iterate)
        expected_dir = np.array([-0.5, -0.5])
        np.testing.assert_array_almost_equal(the_dir, expected_dir, decimal=6)

        optimal_solution = np.array([0, 0])
        the_dir = inverse_bfgs_direction.get_direction(optimal_solution)
        self.assertIsNone(the_dir)


class LineSearchTestCase(unittest.TestCase):
    def test_linesearch(self):
        fct = MyFunctionToMinimize(dimension=2)
        iterate = np.array([1.0, 1.0])
        descent_direction = np.array([-1.0, -1.0])
        alpha0 = 1.0
        beta1 = 1.0e-4
        beta2 = 0.99
        lbd = 2.0

        alpha = line_search(fct, iterate, descent_direction, alpha0, beta1, beta2, lbd)

        # Perform assertions to validate the output
        self.assertIsInstance(alpha, float)
        self.assertEqual(alpha, 1)

    def test_linesearch_wolfe_1(self):
        fct = MyFunctionToMinimize(dimension=2)
        iterate = np.array([1.0, 1.0])
        descent_direction = np.array([-1.0, -1.0])
        alpha0 = 1.0
        beta1 = 0.98
        beta2 = 0.99
        lbd = 2.0

        alpha = line_search(fct, iterate, descent_direction, alpha0, beta1, beta2, lbd)

        # Perform assertions to validate the output
        self.assertIsInstance(alpha, float)
        self.assertEqual(alpha, 0.03125)

    def test_linesearch_wolfe_2(self):
        fct = MyFunctionToMinimize(dimension=2)
        iterate = np.array([1.0, 1.0])
        descent_direction = np.array([-1.0, -1.0])
        alpha0 = 1.0e-10
        beta1 = 1.0e-5
        beta2 = 1.0e-4
        lbd = 2.0

        alpha = line_search(fct, iterate, descent_direction, alpha0, beta1, beta2, lbd)

        # Perform assertions to validate the output
        self.assertIsInstance(alpha, float)
        self.assertEqual(alpha, 1.7179869184)

    def test_linesearch_lambda_less_than_one(self):
        fct = MyFunctionToMinimize(dimension=2)
        iterate = np.array([1.0, 1.0])
        descent_direction = np.array([-1.0, -1.0])
        alpha0 = 1.0
        beta1 = 1.0e-4
        beta2 = 0.99
        lbd = 0.5

        with self.assertRaises(OptimizationError):
            line_search(fct, iterate, descent_direction, alpha0, beta1, beta2, lbd)

    def test_linesearch_alpha0_less_than_zero(self):
        fct = MyFunctionToMinimize(dimension=2)
        iterate = np.array([1.0, 1.0])
        descent_direction = np.array([-1.0, -1.0])
        alpha0 = -1.0
        beta1 = 1.0e-4
        beta2 = 0.99
        lbd = 2.0

        with self.assertRaises(OptimizationError):
            line_search(fct, iterate, descent_direction, alpha0, beta1, beta2, lbd)

    def test_linesearch_beta1_greater_than_beta2(self):
        fct = MyFunctionToMinimize(dimension=2)
        iterate = np.array([1.0, 1.0])
        descent_direction = np.array([-1.0, -1.0])
        alpha0 = 1.0
        beta1 = 0.99
        beta2 = 0.5
        lbd = 2.0

        with self.assertRaises(OptimizationError):
            line_search(fct, iterate, descent_direction, alpha0, beta1, beta2, lbd)

    def test_linesearch_non_descent_direction(self):
        fct = MyFunctionToMinimize(dimension=2)
        iterate = np.array([1.0, 1.0])
        descent_direction = np.array([1.0, 1.0])
        alpha0 = 1.0
        beta1 = 1.0e-4
        beta2 = 0.99
        lbd = 2.0

        with self.assertRaises(OptimizationError):
            line_search(fct, iterate, descent_direction, alpha0, beta1, beta2, lbd)

    def test_newton_linesearch(self):
        fct = MyFunctionToMinimize(dimension=2)

        # Set the starting point
        starting_point = np.array([2.0, 2.0])

        # Set the expected solution
        expected_solution = np.array([0.0, 0.0])

        # Call the newton_linesearch function
        optimization_results = newton_line_search(
            the_function=fct, starting_point=starting_point
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 0 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

    def test_bfgs_linesearch(self):
        fct = MyFunctionToMinimize(dimension=2)

        # Set the starting point
        starting_point = np.array([2.0, 2.0])

        # Set the expected solution
        expected_solution = np.array([0.0, 0.0])

        # Call the newton_linesearch function
        optimization_results = bfgs_line_search(
            the_function=fct, starting_point=starting_point
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 0 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

    def test_newton_linesearch_ex58(self):
        fct = Example58()

        # Set the starting point
        starting_point = np.array([1.0, 1.0])

        # Set the expected solution
        expected_solution = np.array([1.0, np.pi])

        # Call the newton_linesearch function
        optimization_results = newton_line_search(
            the_function=fct, starting_point=starting_point
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 2.1e-10 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

    def test_bfgs_linesearch_58(self):
        fct = Example58()

        # Set the starting point
        starting_point = np.array([1.0, 1.0])

        # Set the expected solution
        expected_solution = np.array([-1.0, 0])

        # Call the newton_linesearch function
        optimization_results = bfgs_line_search(
            the_function=fct, starting_point=starting_point
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-3
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 3.4e-06 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

    def test_newton_linesearch_rosenbrock(self):
        fct = Rosenbrock()

        # Set the starting point
        starting_point = np.array([-1.5, 1.5])

        # Set the expected solution
        expected_solution = np.array([1.0, 1.0])

        # Call the newton_linesearch function
        optimization_results = newton_line_search(
            the_function=fct, starting_point=starting_point
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 2.1e-07 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

    def test_bfgs_linesearch_rosenbrock(self):
        fct = Rosenbrock()

        # Set the starting point
        starting_point = np.array([-1.5, 1.5])

        # Set the expected solution
        expected_solution = np.array([1.0, 1.00])

        # Call the newton_linesearch function
        optimization_results = bfgs_line_search(
            the_function=fct, starting_point=starting_point
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-3
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 1.5e-07 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)


if __name__ == '__main__':
    unittest.main()



"""File diagnostics.py

:author: Michel Bierlaire
:date: Wed Jun 28 16:17:43 2023

Classes for algorithm diagnostics
"""

import numpy as np
from enum import Enum, auto
from typing import NamedTuple


class OptimizationResults(NamedTuple):
    solution: np.ndarray
    messages: dict[str, str]
    convergence: bool


class Diagnostic(Enum):
    pass


class DoglegDiagnostic(Diagnostic):
    """Possible outcomes of the dogleg method"""

    NEGATIVE_CURVATURE = auto()
    CAUCHY = auto()
    PARTIAL_CAUCHY = auto()
    NEWTON = auto()
    PARTIAL_NEWTON = auto()
    DOGLEG = auto()


class ConjugateGradientDiagnostic(Diagnostic):
    """Possible outcomes of the conjugate gradient method"""

    CONVERGENCE = auto()
    OUT_OF_TRUST_REGION = auto()
    NEGATIVE_CURVATURE = auto()
    NUMERICAL_PROBLEM = auto()



"""File linesearch.py

:author: Michel Bierlaire
:date: Thu Jun 22 15:12:21 2023

Functions for line search algorithms
"""

from abc import ABC, abstractmethod
import logging
import numpy as np
from biogeme_optimization.diagnostics import OptimizationResults
from biogeme_optimization.exceptions import OptimizationError
from biogeme_optimization.bfgs import inverse_bfgs
from biogeme_optimization.algebra import schnabel_eskow_direction
from biogeme_optimization.function import FunctionToMinimize

logger: logging.Logger = logging.getLogger(__name__)


class DirectionManagement(ABC):
    """Abstract class for the generation of a descent direction."""

    def __init__(self, function: FunctionToMinimize):
        """Constructor

        :param function: the function to minimize

        """
        self.the_function = function

    @abstractmethod
    def get_direction(self, iterate: np.ndarray) -> np.ndarray | None:
        """Obtain a descent direction

        :param iterate: current iterate
        :type iterate: numpy.array (n x 1)

        """


class NewtonDirection(DirectionManagement):
    """Class for the generation of the Newton direction."""

    def get_direction(self, iterate: np.ndarray) -> np.ndarray | None:
        """Update the data necessary to calculate the direction.

        :param iterate: current iterate
        :type iterate: numpy.array (n x 1)

        :return: descent direction, or None if the current iterate is optimal
        :rtype: numpy.array (n x 1)
        """
        self.the_function.set_variables(iterate)
        evaluation = self.the_function.f_g_h()
        optimal = self.the_function.check_optimality()
        if optimal:
            return None
        direction = schnabel_eskow_direction(evaluation.gradient, evaluation.hessian)
        return direction


class InverseBfgsDirection(DirectionManagement):
    """Class for the generation of the inverse BFGS direction."""

    def __init__(
        self,
        function: FunctionToMinimize,
        first_inverse_approximation: np.ndarray | None = None,
    ) -> None:
        """Constructor

        :param first_inverse_approximation: first inverse approximation of the Hessian
        :type first_inverse_approximation: numpy.array (n x n)
        """
        super().__init__(function)

        if first_inverse_approximation is None:
            self.inverse_hessian_approx = np.identity(self.the_function.dimension())
        else:
            if first_inverse_approximation.ndim != self.the_function.dimension():
                error_msg = (
                    f'Incompatible dimensions: {self.the_function.dimension()} '
                    f'and {first_inverse_approximation.ndim}'
                )
                raise OptimizationError(error_msg)
            self.inverse_hessian_approx = first_inverse_approximation
        self.last_iterate = None
        self.last_gradient = None

    def get_direction(self, iterate: np.ndarray) -> np.ndarray | None:
        """Obtain a descent direction

        :param iterate: current iterate
        :type iterate: numpy.array (n x 1)

        :return: descent direction, or None if the current iterate is optimal.
        """
        self.the_function.set_variables(iterate)
        evaluation = self.the_function.f_g()
        optimal = self.the_function.check_optimality()
        if optimal:
            return None
        # The first time, there is nothing to update
        if self.last_iterate is not None:
            delta_x = iterate - self.last_iterate
            delta_g = evaluation.gradient - self.last_gradient
            # Update the approximation, if possible
            try:
                self.inverse_hessian_approx = inverse_bfgs(
                    self.inverse_hessian_approx, delta_x, delta_g
                )
            except OptimizationError as e:
                logger.warning(e)
        self.last_iterate = iterate
        self.last_gradient = evaluation.gradient

        # Calculate and return the direction
        direction = -self.inverse_hessian_approx @ evaluation.gradient
        return direction


def line_search(
    fct: FunctionToMinimize,
    iterate: np.ndarray,
    descent_direction: np.ndarray,
    alpha0: float = 1.0,
    beta1: float = 1.0e-4,
    beta2: float = 0.99,
    lbd: float = 2.0,
    maxiter: int = 1000,
) -> float:
    """
    Calculate a step along a direction that satisfies both Wolfe conditions

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: function_to_minimize

    :param iterate: current iterate.
    :type iterate: numpy.array

    :param descent_direction: descent direction.
    :type descent_direction: numpy.array

    :param alpha0: first step to test.
    :type alpha0: float

    :param beta1: parameter of the first Wolfe condition.
    :type beta1: float

    :param beta2: parameter of the second Wolfe condition.
    :type beta2: float

    :param lbd: expansion factor for a short step.
    :type lbd: float

    :param maxiter: the algorithm stops if this number of iterations
        is reached. Default: 1000

    :return: a step verifying both Wolfe conditions
    :rtype: float

    :raises OptimizationError: if ``lbd`` :math:`\\leq` 1
    :raises OptimizationError: if ``alpha0`` :math:`\\leq` 0
    :raises OptimizationError: if ``beta1`` :math:`\\geq` beta2
    :raises OptimizationError: if ``descent_direction`` is not a descent
                                             direction

    """
    if lbd <= 1:
        raise OptimizationError(f'lambda is {lbd} and must be > 1')
    if alpha0 <= 0:
        raise OptimizationError(f'alpha0 is {alpha0} and must be > 0')
    if beta1 >= beta2:
        error_msg = (
            f'Incompatible Wolfe cond. parameters: '
            f'beta1= {beta1} is greater than '
            f'beta2={beta2}'
        )
        raise OptimizationError(error_msg)

    fct.set_variables(iterate)
    evaluation = fct.f_g()
    deriv = np.inner(evaluation.gradient, descent_direction)

    if deriv >= 0:
        raise OptimizationError(
            f'descent_direction is not a descent direction: {deriv} >= 0'
        )

    alpha = alpha0
    alpha_left = 0
    alpha_right = np.inf
    for _ in range(maxiter):
        candidate = iterate + alpha * descent_direction
        fct.set_variables(candidate)
        candidate_evaluation = fct.f_g()
        # First Wolfe condition violated?
        if candidate_evaluation.function > evaluation.function + alpha * beta1 * deriv:
            alpha_right = alpha
            alpha = (alpha_left + alpha_right) / 2.0
        # Second Wolfe condition violated?
        elif np.inner(candidate_evaluation.gradient, descent_direction) < beta2 * deriv:
            alpha_left = alpha
            if alpha_right == np.inf:
                alpha *= lbd
            else:
                alpha = (alpha_left + alpha_right) / 2.0
        else:
            return alpha
    else:
        raise OptimizationError(
            f'Line search algorithm could not find a step verifying both Wolfe '
            f'conditions after {maxiter} iterations.'
        )


def minimization_with_line_search(
    the_function: FunctionToMinimize,
    starting_point: np.ndarray,
    direction_management: DirectionManagement,
    maxiter: int = 1000,
) -> OptimizationResults:
    """Minimization method with inexact line search (Wolfe conditions)

    :param the_function: function to minimize
    :type the_function: FunctionToMinimize

    :param starting_point: starting point
    :type starting_point: numpy.array

    :param direction_management: object in charge of providing the
        descent directions and to check optimality
    :type direction_management: DirectionManagement

    :param maxiter: the algorithm stops if this number of iterations
                    is reached. Default: 1000
    :type maxiter: int

    :return: named tuple:

            - solution is the solution found,
            - message is a dictionary reporting various aspects
              related to the run of the algorithm.
            - convergence is a bool which is True if the algorithm has converged

    :rtype: OptimizationResults

    :raises OptimizationError:

    """
    xk = starting_point

    for k in range(maxiter):
        direction = direction_management.get_direction(xk)
        if direction is None:
            # The current iterate is optimal
            messages = the_function.messages
            messages['Number of iterations'] = k
            optimization_results = OptimizationResults(
                solution=xk,
                messages=messages,
                convergence=True,
            )
            break
        alpha = line_search(the_function, xk, direction)
        delta = alpha * direction
        xk = xk + delta
    else:
        messages = the_function.messages
        messages['Cause of termination'] = (
            f'Maximum number of iterations reached: {maxiter}'
        )

        optimization_results = OptimizationResults(
            solution=xk,
            messages=messages,
            convergence=False,
        )
    return optimization_results


def newton_line_search(
    *,
    the_function: FunctionToMinimize,
    starting_point: np.ndarray,
    maxiter: int = 1000,
) -> OptimizationResults:
    """
    Newton method with inexact line search (Wolfe conditions)

    :param the_function: object to calculate the objective function and its derivatives.
    :type the_function: optimization.functionToMinimize

    :param starting_point: starting point
    :type starting_point: numpy.array

    :param maxiter: the algorithm stops if this number of iterations
                    is reached. Default: 1000
    :type maxiter: int

    :return: named tuple:

            - solution is the solution found,
            - message is a dictionary reporting various aspects
              related to the run of the algorithm.
            - convergence is a bool which is True if the algorithm has converged

    :rtype: OptimizationResults

    """
    messages = {'Algorithm': 'Unconstrained Newton with line search'}
    newton_direction = NewtonDirection(the_function)
    optimization_results = minimization_with_line_search(
        the_function, starting_point, newton_direction, maxiter
    )
    messages.update(optimization_results.messages)
    return OptimizationResults(
        solution=optimization_results.solution,
        messages=messages,
        convergence=optimization_results.convergence,
    )


def bfgs_line_search(
    *,
    the_function: FunctionToMinimize,
    starting_point: np.ndarray,
    init_bfgs: np.ndarray | None = None,
    maxiter: int = 1000,
) -> OptimizationResults:
    """BFGS method with inexact line search (Wolfe conditions)

    :param the_function: object to calculate the objective function and its derivatives.
    :type the_function: optimization.functionToMinimize

    :param starting_point: starting point
    :type starting_point: numpy.array

    :param init_bfgs: matrix used to initialize BFGS. If None, the
                     identity matrix is used. Default: None.
    :type init_bfgs: numpy.array

    :param maxiter: the algorithm stops if this number of iterations
                    is reached. Default: 1000
    :type maxiter: int

    :return: named tuple:

            - solution is the solution found,
            - message is a dictionary reporting various aspects
              related to the run of the algorithm.
            - convergence is a bool which is True if the algorithm has converged

    :rtype: OptimizationResult

    """
    messages = {'Algorithm': 'Inverse BFGS with line search'}
    bfgs_direction = InverseBfgsDirection(
        function=the_function, first_inverse_approximation=init_bfgs
    )
    optimization_results = minimization_with_line_search(
        the_function, starting_point, bfgs_direction, maxiter
    )
    messages.update(optimization_results.messages)
    return OptimizationResults(
        solution=optimization_results.solution,
        messages=messages,
        convergence=optimization_results.convergence,
    )



""" Interface for the use of the loggers in the module

:author: Michel Bierlaire
:date: Fri Jul  7 15:18:07 2023

"""

import logging

DEBUG: int = logging.DEBUG
INFO: int = logging.INFO
WARNING: int = logging.WARNING
ERROR: int = logging.ERROR
CRITICAL: int = logging.CRITICAL

loggers: tuple[logging.Logger, ...] = (logging.getLogger('biogeme_optimization'),)


def get_screen_logger(level: int = WARNING) -> logging.Logger:
    """Obtain a screen logger

    :param level: level of verbosity of the logger
    :type level: int
    """
    for logger in loggers:
        # This has to be set to the lower level: DEBUG so that it does not
        # superseed the handler
        logger.setLevel(DEBUG)
        formatter_debug = logging.Formatter(
            '[%(levelname)s] ' '%(asctime)s ' '%(message)s ' '<%(filename)s:%(lineno)d>'
        )
        formatter_normal = logging.Formatter('%(message)s ')
        formatter = formatter_debug if level == DEBUG else formatter_normal
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(level)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
    return loggers[0]


def get_file_logger(filename: str, level: int = WARNING) -> logging.Logger:
    """Obtain a file logger

    :param filename: name of the file. Extension .log recommended.
    :type filename: str

    :param level: level of verbosity of the logger
    :type level: int
    """
    for logger in loggers:
        # This has to be set to the lower level: DEBUG so that it does not
        # superseed the handler
        logger.setLevel(DEBUG)
        formatter = logging.Formatter(
            '[%(levelname)s] ' '%(asctime)s ' '%(message)s ' '<%(filename)s:%(lineno)d>'
        )
        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    return loggers[0]



"""File simple_bounds.py

:author: Michel Bierlaire
:date: Wed Jun 28 11:40:55 2023

Functions for the trust region algorithm with simple bounds
"""

import logging
import numpy as np

from biogeme_optimization.bounds import Bounds
from biogeme_optimization.exceptions import OptimizationError
from biogeme_optimization.function import FunctionToMinimize
from biogeme_optimization.hybrid_function import HybridFunction
from biogeme_optimization.format import Column, FormattedColumns
from biogeme_optimization.diagnostics import OptimizationResults

logger = logging.getLogger(__name__)


def simple_bounds_newton_algorithm(
    *,
    the_function: FunctionToMinimize,
    bounds: Bounds,
    starting_point: np.ndarray,
    variable_names: list[str] | None = None,
    proportion_analytical_hessian: float = 1.0,
    first_radius: float = 1.0,
    conjugate_gradient_tol: float = np.finfo(np.float64).eps ** 0.3333,
    maxiter: int = 1000,
    eta1: float = 0.01,
    eta2: float = 0.9,
    enlarging_factor: float = 10.0,
) -> OptimizationResults:
    """Trust region algorithm for problems with simple bounds

    :param the_function: object to calculate the objective function and its derivatives.
    :type the_function: optimization.functionToMinimize

    :param bounds: bounds on the unsorted_set_of_variables
    :type bounds: class Bounds

    :param starting_point: starting point
    :type starting_point: numpy.array

    :param variable_names: names of the unsorted_set_of_variables, for reporting purposes
    :type variable_names: list(str)

    :param proportion_analytical_hessian: proportion of the iterations where
                                  the true hessian is calculated. When
                                  not, the BFGS update is used. If
                                  1.0, it is used for all
                                  iterations. If 0.0, it is not used
                                  at all.
    :type proportion_analytical_hessian: float

    :param first_radius: initial radius of the trust region. Default: 100.
    :type first_radius: float

    :param conjugate_gradient_tol: the conjugate gradient algorithm stops when this
                  precision is reached.  Default:
                  :math:`\\varepsilon^{\\frac{1}{3}}`

    :type conjugate_gradient_tol: float

    :param maxiter: the algorithm stops if this number of iterations
                    is reached. Default: 1000.

    :type maxiter: int

    :param eta1: threshold for failed iterations. Default: 0.01.
    :type eta1: float

    :param eta2: threshold for very successful iterations. Default 0.9.
    :type eta2: float

    :param enlarging_factor: if an iteration is very successful, the
                            radius of the trust region is multiplied
                            by this factor. Default 10.
    :type enlarging_factor: float

    :return: named tuple:

            - solution is the solution found,
            - message is a dictionary reporting various aspects
              related to the run of the algorithm.
            - convergence is a bool which is True if the algorithm has converged

    :rtype: OptimizationResults

    :raises biogeme.exceptions.BiogemeError: if the dimensions of the
             matrix initBfgs do not match the length of starting_point.

    """
    dimension = bounds.dimension
    if len(starting_point) != dimension:
        raise OptimizationError(
            f'Incompatible size:' f' {len(starting_point)} and {dimension}'
        )

    if not bounds.feasible(starting_point):
        logger.warning(
            'Initial point not feasible. '
            'It will be projected onto the feasible domain.'
        )

    if the_function.needs_reset():
        warning_msg = (
            'It seems that the function has been previously used in another '
            'context, as previously calculated values are stored. It will not '
            'have any consequence on the algorithm. It will just modify the '
            'statistics about function evaluation, as these previous calculations '
            'will be accounted for. It is was not intended, call the method '
            'reset() of the function before using it.'
        )
        logger.warning(warning_msg)

    # Establish the name of the algorithm, for reporting

    if proportion_analytical_hessian < 0 or proportion_analytical_hessian > 1:
        error_msg = (
            'Proportion must be between 0 and 1, not {proportion_analytical_hessian}'
        )
        raise OptimizationError(error_msg)

    if proportion_analytical_hessian == 1.0:
        algo = 'Newton with trust region for simple bound constraints'
    elif proportion_analytical_hessian == 0.0:
        algo = 'BFGS with trust region for simple bound constraints'
    else:
        algo = (
            f'Hybrid Newton [{100*proportion_analytical_hessian}%] with trust '
            f'region for simple bound constraints'
        )

    columns = [Column(title='Iter.', width=5)]

    if variable_names is not None:
        columns += [Column(title=name[:15], width=15) for name in variable_names]
    columns += [
        Column(title='Function', width=12),
        Column(title='Relgrad', width=10),
        Column(title='Radius', width=8),
        Column(title='Rho', width=8),
        Column(title=' ', width=4),
    ]

    the_formatter = FormattedColumns(columns)

    the_hybrid_function = HybridFunction(
        the_function, proportion_analytical_hessian, bounds=bounds
    )

    iterate = bounds.project(starting_point)
    current_function = the_hybrid_function.calculate_function_and_derivatives(iterate)
    if current_function is None:
        # The starting point is optimal
        messages = the_function.messages
        messages['Algorithm'] = algo
        messages['Number of iterations'] = '0'
        return OptimizationResults(
            solution=iterate, messages=messages, convergence=True
        )

    radius = first_radius
    max_delta = 1.0e10
    min_delta = np.sqrt(np.finfo(float).eps)
    rho = 0.0
    status = ''

    logger.info(the_formatter.formatted_title())

    for k in range(maxiter):

        def logmessage() -> None:
            """Send messages to the logger"""
            values_to_report = [k]
            if variable_names is not None:
                values_to_report += list(iterate)
            values_to_report += [
                current_function.function,
                the_function.relative_gradient_norm,
                float(radius),
                rho,
                status,
            ]
            logger.info(the_formatter.formatted_row(values_to_report))

        if (
            np.isnan(current_function.function).any()
            or np.isnan(current_function.gradient).any()
            or np.isnan(current_function.hessian).any()
        ):
            radius /= 2.0
            if radius <= min_delta:
                messages = the_function.messages
                messages['Algorithm'] = algo
                message = f'Trust region is too small: {radius}'
                messages['Cause of termination'] = message
                messages['Number of iterations'] = f'{k + 1}'
                messages['Proportion of Hessian calculation'] = (
                    the_hybrid_function.message()
                )
                logmessage()
                return OptimizationResults(
                    solution=iterate, messages=messages, convergence=False
                )

            status = '-'
            logmessage()
            continue
        try:
            # Solve the quadratic problem in the subspace defined by the GCP
            candidate, _ = bounds.truncated_conjugate_gradient_subspace(
                iterate=iterate,
                gradient=current_function.gradient,
                hessian=current_function.hessian,
                radius=radius,
                tol=conjugate_gradient_tol,
            )
            if np.isnan(candidate).any():
                radius /= 2.0
                if radius <= min_delta:
                    messages = the_function.messages
                    messages['Algorithm'] = algo
                    message = f'Trust region is too small: {radius}'
                    messages['Cause of termination'] = message
                    messages['Number of iterations'] = f'{k+1}'
                    messages['Proportion of Hessian calculation'] = (
                        the_hybrid_function.message()
                    )
                    logmessage()
                    return OptimizationResults(
                        solution=iterate, messages=messages, convergence=False
                    )

                status = '-'
                logmessage()
                continue

            try:
                function_value_candidate = the_hybrid_function.calculate_function(
                    candidate
                )
                failed = not np.isfinite(function_value_candidate)
                step = None
                if not failed:
                    num = current_function.function - function_value_candidate
                    step = candidate - iterate
                    denom = -np.inner(step, current_function.gradient) - 0.5 * np.inner(
                        step, current_function.hessian @ step
                    )
                    rho = num / denom
                    failed = rho < eta1
                if failed:
                    if step is None:
                        radius /= 2.0
                    else:
                        radius = min(radius / 2.0, np.linalg.norm(step, np.inf) / 2.0)
                    if radius <= min_delta:
                        messages = the_function.messages
                        messages['Algorithm'] = algo
                        message = f'Trust region is too small: {radius}'
                        messages['Cause of termination'] = message
                        messages['Number of iterations'] = f'{k+1}'
                        messages['Proportion of Hessian calculation'] = (
                            the_hybrid_function.message()
                        )
                        logmessage()
                        return OptimizationResults(
                            solution=iterate, messages=messages, convergence=False
                        )

                    status = '-'
                    logmessage()
                    continue
            except RuntimeError as e:
                logger.warning(e)
                radius /= 2.0
                if radius <= min_delta:
                    messages = the_function.messages
                    messages['Algorithm'] = algo
                    message = f'Trust region is too small: {radius}'
                    messages['Cause of termination'] = message
                    messages['Number of iterations'] = f'{k+1}'
                    messages['Proportion of Hessian calculation'] = (
                        the_hybrid_function.message()
                    )
                    logmessage()
                    return OptimizationResults(
                        solution=iterate, messages=messages, convergence=False
                    )
                failed = True
                status = '-'
                logmessage()
                continue

            # Candidate accepted
            try:
                candidate_function = (
                    the_hybrid_function.calculate_function_and_derivatives(candidate)
                )
            except OptimizationError:
                # Failure: reduce the trust region
                radius = min(radius / 2.0, np.linalg.norm(step, np.inf) / 2.0)
                if radius <= min_delta:
                    messages = the_function.messages
                    messages['Algorithm'] = algo
                    message = f'Trust region is too small: {radius}'
                    messages['Cause of termination'] = message
                    messages['Number of iterations'] = f'{k+1}'
                    messages['Proportion of Hessian calculation'] = (
                        the_hybrid_function.message()
                    )
                    logmessage()
                    return OptimizationResults(
                        solution=iterate, messages=messages, convergence=False
                    )
                status = '-'
                logmessage()

                continue

            if candidate_function is None:
                # the_matrix stopping criterion has been detected
                messages = the_function.messages
                messages['Algorithm'] = algo
                messages['Number of iterations'] = f'{k+1}'
                messages['Proportion of Hessian calculation'] = (
                    the_hybrid_function.message()
                )
                logmessage()
                return OptimizationResults(
                    solution=candidate, messages=messages, convergence=True
                )

            iterate = candidate
            current_function = candidate_function

            if rho >= eta2:
                # Enlarge the trust region
                radius = min(enlarging_factor * radius, max_delta)
                status = '++'
            else:
                status = '+'

            logmessage()
        except KeyboardInterrupt:
            warning_msg = (
                'You have pressed Ctrl-C. The iterations of the optimization algorithms '
                'are interrupted, but the script will continue to run. If you want to interrupt the script, '
                'press Ctrl-C again'
            )
            logger.warning(warning_msg)
            messages = the_function.messages
            messages['Algorithm'] = algo
            messages['Number of iterations'] = f'{k + 1}'
            message = 'Iterations interrupted by the user'
            messages['Cause of termination'] = message
            messages['Proportion of Hessian calculation'] = (
                the_hybrid_function.message()
            )
            logmessage()
            return OptimizationResults(
                solution=iterate, messages=messages, convergence=False
            )

    the_function.calculate_relative_projected_gradient(bounds)
    messages = the_function.messages
    messages['Algorithm'] = algo
    message = f'Maximum number of iterations reached: {maxiter}'
    messages['Cause of termination'] = message
    messages['Number of iterations'] = f'{maxiter}'
    messages['Proportion of Hessian calculation'] = the_hybrid_function.message()
    return OptimizationResults(solution=iterate, messages=messages, convergence=False)



"""Defines a generic exception for the optimization package

:author: Michel Bierlaire

:date: Wed Jun 21 10:34:51 2023

"""


class OptimizationError(Exception):
    """Defines a generic exception ."""



"""File test_function.py

:author: Michel Bierlaire
:date: Thu Jun 22 17:08:01 2023

Tests for the function module
"""

import unittest
import numpy as np

from biogeme_optimization.function import relative_gradient
from examples import MyFunctionToMinimize, Example58, Rosenbrock


class RelativeGradientTestCase(unittest.TestCase):
    def test_relative_gradient(self):
        x = np.array([1.0, 2.0, 3.0])
        f = 10.0
        g = np.array([0.1, 0.2, 0.3])
        typx = np.array([0.5, 1.0, 1.5])
        typf = 5.0

        result = relative_gradient(x, f, g, typx, typf)

        # Expected relative gradients for each component
        expected_relgrad = np.array([0.2, 0.4, 0.6])
        # Expected maximum relative gradient
        expected_result = 0.09

        # Perform assertions to validate the output
        np.testing.assert_allclose(result, expected_result, rtol=1e-6)

    def test_relative_gradient_finite_result(self):
        x = np.array([1.0, 2.0, 3.0])
        f = 10.0
        g = np.array([0.1, 0.2, 0.3])
        typx = np.array([0.5, 1.0, 1.5])
        typf = 5.0

        # Set one component of g to infinity
        g[1] = np.inf

        result = relative_gradient(x, f, g, typx, typf)

        # Expected result when one component of g is infinite
        expected_result = np.finfo(float).max

        # Perform assertions to validate the output
        self.assertEqual(result, expected_result)


class TestFiniteDifferencesGradient(unittest.TestCase):
    def test_finite_differences_gradient(self):

        instances = (MyFunctionToMinimize(dimension=2), Example58(), Rosenbrock())

        for instance in instances:
            # Test case
            x = np.array([1, 2])
            instance.set_variables(x)
            evaluation = instance.f_g()
            expected_gradient = evaluation.gradient

            # Call the function being tested
            fin_diff_gradient = instance.finite_differences_gradient(x)

            # Assertion
            np.testing.assert_array_almost_equal(
                expected_gradient, fin_diff_gradient, decimal=3
            )


class TestFiniteDifferencesHessian(unittest.TestCase):
    def test_finite_differences_hessian(self):
        # Create a mock class that mimics the behavior of the original class
        instances = (MyFunctionToMinimize(dimension=2), Example58(), Rosenbrock())

        for instance in instances:
            # Test case
            x = np.array([1, 2])
            instance.set_variables(x)
            evaluation = instance.f_g_h()
            expected_hessian = evaluation.hessian

            # Call the function being tested
            fin_diff_hessian = instance.finite_differences_hessian(x)

            # Assertion
            np.testing.assert_array_almost_equal(
                expected_hessian, fin_diff_hessian, decimal=3
            )


if __name__ == '__main__':
    unittest.main()



"""File function.py

:author: Michel Bierlaire
:date: Thu Jun 22 16:42:46 2023

Abstract class for the function to minimize
"""

import logging
from abc import ABC, abstractmethod
from typing import final, NamedTuple

import numpy as np

from biogeme_optimization.bounds import Bounds


class FunctionData(NamedTuple):
    function: float | None
    gradient: np.ndarray | None
    hessian: np.ndarray | None


class CheckDerivativesResults(NamedTuple):
    analytical: FunctionData
    finite_differences: FunctionData


logger: logging.Logger = logging.getLogger(__name__)

SQRT_EPSILON: float = np.sqrt(np.finfo(np.float64).eps)
EPSILON: float = float(np.finfo(np.float64).eps)


class FunctionToMinimize(ABC):
    """This is an abstract class. The actual function to minimize
    must be implemented in a concrete class deriving from this one.

    """

    def __init__(
        self,
        epsilon: float | None = None,
        steptol: float | None = None,
    ) -> None:
        """Constructor
        :param epsilon: tolerance on the relative gradient for convergence
        :type epsilon: float

        :param steptol: tolerance of the step length. If the distance
            between two successive iterates is less than steptol, the
            iterations should be interrupted.
        :type  steptol: float

        """
        self.epsilon = (
            np.finfo(np.float64).eps ** 0.3333 if epsilon is None else epsilon
        )
        self.steptol = 1.0e-5 if steptol is None else steptol
        self.reset()

    @final
    def reset(self) -> None:
        """Erase all stored values"""
        self.stored_values: dict[bytes, float] = {}
        self.stored_gradient: dict[bytes, FunctionData] = {}
        self.stored_hessian: dict[bytes, FunctionData] = {}
        self.x: np.ndarray | None = None
        self.x_bytes: bytes | None = None
        self.typx: np.ndarray | None = None
        self.typf: float | None = None
        self.relative_gradient_norm: float | None = None
        self.messages: dict[str, str | float | int] = {}

    @final
    def needs_reset(self) -> bool:
        """Checks if the function needs to be reset

        :return: True if calculated values of the function are stored
        :rtype: bool
        """
        return bool(self.stored_values)

    @abstractmethod
    def dimension(self) -> int:
        """Provides the number of unsorted_set_of_variables of the problem"""

    def calculate_relative_projected_gradient(self, bounds: Bounds = None) -> None:
        """Calculates the relative projected gradient

        :param bounds: object describing the bound constraints
        :type bounds: Bounds

        """
        evaluation = self.f_g()
        value_function = evaluation.function
        gradient = evaluation.gradient

        projected_gradient = (
            gradient if bounds is None else bounds.project(self.x - gradient) - self.x
        )

        if self.typx is None:
            self.typx = np.ones(np.asarray(self.x).shape)
        if self.typf is None:
            self.typf = max(np.abs(evaluation.function), 1.0)
        self.relative_gradient_norm = relative_gradient(
            self.x, value_function, projected_gradient, self.typx, self.typf
        )

    def check_optimality(self, bounds: Bounds = None) -> bool:
        """Verifies the optimality of the current iterate

        :param bounds: object describing the bound constraints
        :type bounds: Bounds

        :return: True is optimal, False otherwise
        :rtype: bool

        """
        self.calculate_relative_projected_gradient(bounds)
        if self.relative_gradient_norm <= self.epsilon:
            message = f'Relative gradient = {self.relative_gradient_norm:.2g} <= {self.epsilon:.2g}'
            self.messages['Relative gradient'] = self.relative_gradient_norm
            self.messages['Cause of termination'] = message
            self.messages['Number of function evaluations'] = (
                self.nbr_function_evaluations()
            )
            self.messages['Number of gradient evaluations'] = (
                self.nbr_gradient_evaluations()
            )
            self.messages['Number of hessian evaluations'] = (
                self.nbr_hessian_evaluations()
            )
            return True
        return False

    def check_insufficient_progress(self, previous_iterate: np.ndarray) -> bool:
        """Verifies if the algorithm is making sufficient progress. If
            not, the algorithm is stopped

        :param previous_iterate: previous iterate
        :type previous_iterate: numpy.array (n x 1)

        :return: True if progress is insufficient, False otherwise
        :rtype: bool

        """
        if self.typx is None:
            self.typx = np.ones(np.asarray(self.x).shape)
        change = relative_change(self.x, previous_iterate, self.typx)
        if change <= self.steptol:
            message = f'Relative change = ' f'{change:.3g} <= {self.steptol:.2g}'
            self.messages['Cause of termination'] = message
            self.messages['Number of function evaluations'] = (
                self.nbr_function_evaluations()
            )
            self.messages['Number of gradient evaluations'] = (
                self.nbr_gradient_evaluations()
            )
            self.messages['Number of hessian evaluations'] = (
                self.nbr_hessian_evaluations()
            )
            return True
        return False

    @final
    def set_variables(self, x: np.ndarray) -> None:
        """Set the values of the unsorted_set_of_variables for which the function
        has to be calculated.

        :param x: values
        :type x: numpy.array
        """
        self.x = x
        self.x_bytes = x.tobytes()

    @final
    def f(self) -> float:
        """Retrieve or calculate the canonical_value of the function

        :return: canonical_value of the function
        :rtype: float
        """
        value = self.stored_values.get(self.x_bytes)
        if value is None:
            value = self._f()
            self.stored_values[self.x_bytes] = value
        return value

    @abstractmethod
    def _f(self) -> float:
        """Calculate the canonical_value of the function

        :return: canonical_value of the function
        :rtype: float
        """

    @final
    def f_g(self) -> FunctionData:
        """Retrieve or calculate the canonical_value of the function and the gradient

        :return: canonical_value of the function and the gradient
        :rtype: FunctionData
        """
        function_data = self.stored_gradient.get(self.x_bytes)
        if function_data is None:
            function_data = self._f_g()
            self.stored_gradient[self.x_bytes] = function_data
            self.stored_values[self.x_bytes] = function_data.function
        return function_data

    @abstractmethod
    def _f_g(self) -> FunctionData:
        """Calculate the canonical_value of the function and the gradient

        :return: canonical_value of the function and the gradient
        :rtype: FunctionData
        """

    @final
    def f_g_h(self) -> FunctionData:
        """Calculate the canonical_value of the function, the gradient and the Hessian

        :return: canonical_value of the function, the gradient and the Hessian
        :rtype: FunctionData
        """
        function_data = self.stored_hessian.get(self.x_bytes)
        if function_data is None:
            function_data = self._f_g_h()
            self.stored_hessian[self.x_bytes] = function_data
            self.stored_gradient[self.x_bytes] = function_data
            self.stored_values[self.x_bytes] = function_data.function
        return function_data

    @abstractmethod
    def _f_g_h(self) -> FunctionData:
        """Calculate the canonical_value of the function, the gradient and the Hessian

        :return: canonical_value of the function, the gradient and the Hessian
        :rtype: FunctionData
        """

    @final
    def nbr_function_evaluations(self) -> int:
        """Obtain the total number of function evaluations"""
        return len(self.stored_values)

    @final
    def nbr_gradient_evaluations(self) -> int:
        """Obtain the total number of gradient evaluations"""
        return len(self.stored_gradient)

    @final
    def nbr_hessian_evaluations(self) -> int:
        """Obtain the total number of hessian evaluations"""
        return len(self.stored_hessian)

    def finite_differences_gradient(self, x: np.ndarray) -> np.ndarray:
        """Calculates the gradient of the function using finite differences

        :param x: argument of the function
        :type x: numpy.array

        :return: numpy vector, same dimension as x, containing the gradient
        calculated by finite differences.
        :rtype: numpy.array

        """
        x = x.astype(float)
        n = len(x)
        g = np.zeros(n)
        tau = SQRT_EPSILON
        self.set_variables(x)
        f = self.f()
        for i in range(n):
            xi = x[i]
            xp = x.copy()
            if abs(xi) >= 1:
                s = tau * xi
            elif xi >= 0:
                s = tau
            else:
                s = -tau
            xp[i] = xi + s
            self.set_variables(xp)
            fp = self.f()
            g[i] = (fp - f) / s
        return g.astype(float)

    def finite_differences_hessian(self, x: np.ndarray) -> np.ndarray:
        """Calculates the hessian of the function using finite differences

        :param x: argument of the function
        :type x: numpy.array

        :return: numpy matrix containing the hessian calculated by
             finite differences.
        :rtype: numpy.array

        """
        tau = 1.0e-7
        n = len(x)
        hessian = np.zeros((n, n))
        self.set_variables(x)
        evaluation = self.f_g()
        g = evaluation.gradient
        eye = np.eye(n)
        for i in range(n):
            xi = x[i]
            if abs(xi) >= 1:
                s = tau * xi
            elif xi >= 0:
                s = tau
            else:
                s = -tau
            ei = eye[i]
            self.set_variables(x + s * ei)
            diff_evaluation = self.f_g()
            gp = diff_evaluation.gradient
            hessian[:, i] = (gp - g) / s
        return hessian.astype(float)

    def check_derivatives(
        self, x: np.ndarray, names: list[str] | None = None
    ) -> CheckDerivativesResults:
        """Verifies the analytical derivatives of a function by comparing
        them with finite difference approximations.

        :param x: arguments of the function
        :type x: numpy.array

        :param names: the names of the entries of x (for reporting).
        :type names: list(string)

        :return: tuple f, g, h, gdiff, hdiff where

          - f is the canonical_value of the function at x,
          - g is the analytical gradient,
          - h is the analytical hessian,
          - gdiff is the difference between the analytical gradient
            and the finite difference approximation
          - hdiff is the difference between the analytical hessian
            and the finite difference approximation

        :rtype: float, numpy.array,numpy.array,  numpy.array,numpy.array

        """
        x = np.array(x, dtype=float)
        self.set_variables(x)
        evaluation = self.f_g_h()
        f = evaluation.function
        g = evaluation.gradient
        h = evaluation.hessian
        g_num = self.finite_differences_gradient(x)
        gdiff = g - g_num
        if names is None:
            names = [f'x[{i}]' for i in range(len(x))]
        logger.info('x\t\tGradient\tFinDiff\t\tDifference')
        for k, v in enumerate(gdiff):
            logger.info(f'{names[k]:15}\t{g[k]:+E}\t{g_num[k]:+E}\t{v:+E}')

        h_num = self.finite_differences_hessian(x)
        hdiff = h - h_num
        logger.info('Row\t\tCol\t\tHessian\tFinDiff\t\tDifference')
        for row in range(len(hdiff)):
            for col in range(len(hdiff)):
                logger.info(
                    f'{names[row]:15}\t{names[col]:15}\t{h[row,col]:+E}\t'
                    f'{h_num[row,col]:+E}\t{hdiff[row,col]:+E}'
                )
        finite_difference = FunctionData(function=None, gradient=g_num, hessian=h_num)
        results = CheckDerivativesResults(
            analytical=evaluation, finite_differences=finite_difference
        )
        return results


def relative_gradient(
    x: np.ndarray, f: float, g: np.ndarray, typx: np.ndarray, typf: float
) -> float:
    """Calculates the relative gradients.

    It is typically used as stopping criterion.

    :param x: current iterate.
    :type x: numpy.array
    :param f: canonical_value of f(x)
    :type f: float
    :param g: :math:`\\nabla f(x)`, gradient of f at x
    :type g: numpy.array
    :param typx: typical canonical_value for x.
    :type typx: numpy.array
    :param typf: typical canonical_value for f.
    :type typf: float

    :return: relative gradient

    .. math:: \\max_{i=1,\\ldots,n}\\frac{(\\nabla f(x))_i
              \\max(x_i,\\text{typx}_i)}
              {\\max(|f(x)|, \\text{typf})}

    :rtype: float
    """
    abs_x_typx = np.maximum(np.abs(x), typx)
    max_abs_f_typf = np.maximum(np.abs(f), typf)

    relgrad = np.abs(g * abs_x_typx / max_abs_f_typf)
    result = np.max(relgrad)

    if np.isfinite(result):
        return result

    return float(np.finfo(float).max)


def relative_change(x: np.ndarray, x_previous: np.ndarray, typx: np.ndarray) -> float:
    """Calculates the relative change.

    It is typically used as stopping criterion.

    :param x: current iterate.
    :type x: numpy.array

    :param x_previous: previous iterate.
    :type x_previous: numpy.array

    :param typx: typical canonical_value for x.
    :type typx: numpy.array

    :return: relative change:

    .. math:: \\max_i \\frac{|(x_k)_i - (x_{k-1})_i|}{\\max(|(x_k)_i|,
              \\text{typx}_i)}

    :rtype: float

    """

    relx = np.array(
        [
            abs(x[i] - x_previous[i]) / np.maximum(abs(x[i]), typx[i])
            for i in range(len(x))
        ]
    )
    result = relx.max()
    if np.isfinite(result):
        return result

    return float(np.finfo(float).max)



"""File test_simple_bounds.py

:author: Michel Bierlaire
:date: Wed Jun 28 11:58:37 2023

Tests for the simple_bounds module
"""
import unittest
import itertools
import numpy as np
from scipy.optimize import minimize
from biogeme_optimization.bounds import Bounds
from biogeme_optimization.simple_bounds import simple_bounds_newton_algorithm

from examples import MyFunctionToMinimize, Rosenbrock


def my_function_for_scipy(x):
    the_function = MyFunctionToMinimize(dimension=3)
    the_function.set_variables(x)
    return the_function.f()


def rosenbrock_for_scipy(x):
    the_function = Rosenbrock()
    the_function.set_variables(x)
    return the_function.f()


class TestSimpleBoundsSimpleQuadratic(unittest.TestCase):
    def test_1(self):
        the_function = MyFunctionToMinimize(dimension=3)
        the_bounds = [(None, None), (None, None), (None, None)]
        bounds = Bounds(the_bounds)
        starting_point = np.array([1, 2, 3])
        values_proportion_analytical_hessian = [0.0, 0.1, 0.5, 1.0]
        values_first_radius = [0.1, 1.0, 10]
        scipy_result = minimize(
            my_function_for_scipy, x0=starting_point, bounds=the_bounds
        )
        expected_solution = scipy_result.x
        for proportion_analytical_hessian, first_radius in itertools.product(
            values_proportion_analytical_hessian, values_first_radius
        ):

            optimization_results = simple_bounds_newton_algorithm(
                the_function=the_function,
                bounds=bounds,
                starting_point=starting_point,
                proportion_analytical_hessian=proportion_analytical_hessian,
                first_radius=first_radius,
            )
            np.testing.assert_array_almost_equal(optimization_results.solution, expected_solution)
            self.assertTrue(optimization_results.convergence)

    def test_2(self):
        the_function = MyFunctionToMinimize(dimension=3)
        the_bounds = [(1, None), (2, None), (3, None)]
        bounds = Bounds(the_bounds)
        starting_point = np.array([10, 20, 30])
        values_proportion_analytical_hessian = [0.0, 0.1, 0.5, 1.0]
        values_first_radius = [0.1, 1.0, 10]
        scipy_result = minimize(
            my_function_for_scipy, x0=starting_point, bounds=the_bounds
        )
        expected_solution = scipy_result.x

        for proportion_analytical_hessian, first_radius in itertools.product(
            values_proportion_analytical_hessian, values_first_radius
        ):

            optimization_results = simple_bounds_newton_algorithm(
                the_function=the_function,
                bounds=bounds,
                starting_point=starting_point,
                proportion_analytical_hessian=proportion_analytical_hessian,
                first_radius=first_radius,
            )
            np.testing.assert_array_almost_equal(optimization_results.solution, expected_solution)
            self.assertTrue(optimization_results.convergence)

    def test_3(self):
        the_function = MyFunctionToMinimize(dimension=3)
        the_bounds = [(-1, 1), (-2, 2), (-3, 3)]
        bounds = Bounds(the_bounds)
        starting_point = np.array([-1, -2, -3])
        values_proportion_analytical_hessian = [0.0, 0.1, 0.5, 1.0]
        values_first_radius = [0.1, 1.0, 10]
        scipy_result = minimize(
            my_function_for_scipy, x0=starting_point, bounds=the_bounds
        )
        expected_solution = scipy_result.x

        for proportion_analytical_hessian, first_radius in itertools.product(
            values_proportion_analytical_hessian, values_first_radius
        ):
            optimization_results = simple_bounds_newton_algorithm(
                the_function=the_function,
                bounds=bounds,
                starting_point=starting_point,
                proportion_analytical_hessian=proportion_analytical_hessian,
                first_radius=first_radius,
            )
            np.testing.assert_array_almost_equal(optimization_results.solution, expected_solution)
            self.assertTrue(optimization_results.convergence)

    def test_4(self):
        the_function = MyFunctionToMinimize(dimension=3)
        the_bounds = [(1, None), (2, None), (3, None)]
        bounds = Bounds(the_bounds)
        starting_point = np.array([10, 2, 30])
        values_proportion_analytical_hessian = [0.0, 0.1, 0.5, 1.0]
        values_first_radius = [0.1, 1.0, 10]
        scipy_result = minimize(
            my_function_for_scipy, x0=starting_point, bounds=the_bounds
        )
        expected_solution = scipy_result.x

        for proportion_analytical_hessian, first_radius in itertools.product(
            values_proportion_analytical_hessian, values_first_radius
        ):

            optimization_results = simple_bounds_newton_algorithm(
                the_function=the_function,
                bounds=bounds,
                starting_point=starting_point,
                proportion_analytical_hessian=proportion_analytical_hessian,
                first_radius=first_radius,
            )
            np.testing.assert_array_almost_equal(optimization_results.solution, expected_solution)
            self.assertTrue(optimization_results.convergence)


class TestSimpleBoundsRosenbrock(unittest.TestCase):
    def test_1(self):
        the_function = Rosenbrock()
        the_bounds = [(None, None), (None, None)]
        bounds = Bounds(the_bounds)
        starting_point = np.array([-1.5, 1.5])
        values_proportion_analytical_hessian = [0.0, 0.1, 0.5, 1.0]
        values_first_radius = [0.1, 1.0, 10]
        scipy_result = minimize(
            rosenbrock_for_scipy, x0=starting_point, bounds=the_bounds
        )
        expected_solution = scipy_result.x
        print(f'{expected_solution=}')
        for proportion_analytical_hessian, first_radius in itertools.product(
            values_proportion_analytical_hessian, values_first_radius
        ):

            optimization_results = simple_bounds_newton_algorithm(
                the_function=the_function,
                bounds=bounds,
                starting_point=starting_point,
                proportion_analytical_hessian=proportion_analytical_hessian,
                first_radius=first_radius,
            )
            np.testing.assert_array_almost_equal(
                optimization_results.solution,
                expected_solution,
                decimal=3
            )
            self.assertTrue(optimization_results.convergence)

    def test_2(self):
        the_function = Rosenbrock()
        the_bounds = [(None, None), (None, 0)]
        bounds = Bounds(the_bounds)
        starting_point = np.array([-1.5, 1.5])
        values_proportion_analytical_hessian = [0.0, 0.1, 0.5, 1.0]
        values_first_radius = [0.1, 1.0, 10]
        scipy_result = minimize(
            rosenbrock_for_scipy, x0=starting_point, bounds=the_bounds
        )
        expected_solution = scipy_result.x

        for proportion_analytical_hessian, first_radius in itertools.product(
            values_proportion_analytical_hessian, values_first_radius
        ):

            optimization_results = simple_bounds_newton_algorithm(
                the_function=the_function,
                bounds=bounds,
                starting_point=starting_point,
                proportion_analytical_hessian=proportion_analytical_hessian,
                first_radius=first_radius,
            )
            np.testing.assert_array_almost_equal(
                optimization_results.solution,
                expected_solution,
                decimal=3
            )
            self.assertTrue(optimization_results.convergence)

    def test_3(self):
        the_function = Rosenbrock()
        the_bounds = [(1, None), (None, 0)]
        bounds = Bounds(the_bounds)
        starting_point = np.array([1, 1.5])
        values_proportion_analytical_hessian = [0.0, 0.1, 0.5, 1.0]
        values_first_radius = [0.1, 1.0, 10]
        scipy_result = minimize(
            rosenbrock_for_scipy, x0=starting_point, bounds=the_bounds
        )
        expected_solution = scipy_result.x

        for proportion_analytical_hessian, first_radius in itertools.product(
            values_proportion_analytical_hessian, values_first_radius
        ):

            optimization_results = simple_bounds_newton_algorithm(
                the_function=the_function,
                bounds=bounds,
                starting_point=starting_point,
                proportion_analytical_hessian=proportion_analytical_hessian,
                first_radius=first_radius,
            )
            np.testing.assert_array_almost_equal(
                optimization_results.solution,
                expected_solution,
                decimal=3
            )
            self.assertTrue(optimization_results.convergence)

    def test_4(self):
        the_function = Rosenbrock()
        the_bounds = [(-12, 1), (-1, 1)]
        bounds = Bounds(the_bounds)
        starting_point = np.array([1, 1.5])
        values_proportion_analytical_hessian = [0.0, 0.1, 0.5, 1.0]
        values_first_radius = [0.1, 1.0, 10]
        scipy_result = minimize(
            rosenbrock_for_scipy, x0=starting_point, bounds=the_bounds
        )
        expected_solution = scipy_result.x

        for proportion_analytical_hessian, first_radius in itertools.product(
            values_proportion_analytical_hessian, values_first_radius
        ):

            optimization_results = simple_bounds_newton_algorithm(
                the_function=the_function,
                bounds=bounds,
                starting_point=starting_point,
                proportion_analytical_hessian=proportion_analytical_hessian,
                first_radius=first_radius,
            )
            np.testing.assert_array_almost_equal(
                optimization_results.solution,
                expected_solution,
                decimal=3
            )
            self.assertTrue(optimization_results.convergence)


if __name__ == '__main__':
    unittest.main()



"""File stochastic_simple_bounds.py

:author: Michel Bierlaire
:date: Fri Aug 11 10:11:00 2023

Functions for the trust region algorithm with simple bounds for stochastic functions
"""

import logging
import numpy as np
from biogeme_optimization.exceptions import OptimizationError
from biogeme_optimization.hybrid_function import HybridFunction
from biogeme_optimization.format import Column, FormattedColumns

logger = logging.getLogger(__name__)


def simple_bounds_newton_algorithm(
    *,
    the_function,
    bounds,
    starting_point,
    variable_names=None,
    proportion_analytical_hessian=1.0,
    first_radius=1.0,
    cgtol=np.finfo(np.float64).eps ** 0.3333,
    maxiter=1000,
    eta1=0.01,
    eta2=0.9,
    enlarging_factor=10,
):
    """Trust region algorithm for problems with simple bounds

    :param the_function: object to calculate the objective function and its derivatives.
    :type the_function: optimization.StochasticFunction

    :param bounds: bounds on the unsorted_set_of_variables
    :type bounds: class Bounds

    :param starting_point: starting point
    :type starting_point: numpy.array

    :param variable_names: names of the unsorted_set_of_variables, for reporting purposes
    :type variable_names: list(str)

    :param proportion_analytical_hessian: proportion of the iterations where
                                  the true hessian is calculated. When
                                  not, the BFGS update is used. If
                                  1.0, it is used for all
                                  iterations. If 0.0, it is not used
                                  at all.
    :type proportion_analytical_hessian: float

    :param first_radius: initial radius of the trust region. Default: 100.
    :type first_radius: float

    :param tol: the algorithm stops when this precision is reached.
                Default: :math:`\\varepsilon^{\\frac{1}{3}}`
    :type tol: float

    :param steptol: the algorithm stops when the relative change in x
                is below this threshold. Basically, if p significant
                digits of x are needed, steptol should be set to
                1.0e-p.  Default: :math:`10^{-5}`

    :type steptol: float


    :param cgtol: the conjugate gradient algorithm stops when this
                  precision is reached.  Default:
                  :math:`\\varepsilon^{\\frac{1}{3}}`

    :type cgtol: float

    :param maxiter: the algorithm stops if this number of iterations
                    is reached. Default: 1000.

    :type maxiter: int

    :param eta1: threshold for failed iterations. Default: 0.01.
    :type eta1: float

    :param eta2: threshold for very successful iterations. Default 0.9.
    :type eta2: float

    :param enlarging_factor: if an iteration is very successful, the
                            radius of the trust region is multiplied
                            by this factor. Default 10.
    :type enlarging_factor: float

    :return: x, messages

        - x is the solution generated by the algorithm,
        - messages is a dictionary describing information about the lagorithm

    :rtype: numpay.array, dict(str:object)

    :raises biogeme.exceptions.BiogemeError: if the dimensions of the
             matrix initBfgs do not match the length of starting_point.

    """
    dimension = bounds.dimension
    if len(starting_point) != dimension:
        raise OptimizationError(
            f'Incompatible size:' f' {len(starting_point)} and {dimension}'
        )

    if not bounds.feasible(starting_point):
        logger.warning(
            'Initial point not feasible. '
            'It will be projected onto the feasible domain.'
        )

    if the_function.needs_reset():
        warning_msg = (
            'It seems that the function has been previously used in another '
            'context, as previously calculated values are stored. It will not '
            'have any consequence on the algorithm. It will just modify the '
            'statistics about function evaluation, as these previous calculations '
            'will be accounted for. It is was not intended, call the method '
            'reset() of the function before using it.'
        )
        logger.warning(warning_msg)

    # Establish the name of the algorithm, for reporting

    if proportion_analytical_hessian < 0 or proportion_analytical_hessian > 1:
        error_msg = (
            'Proportion must be between 0 and 1, not {proportion_analytical_hessian}'
        )
        raise OptimizationError(error_msg)

    if proportion_analytical_hessian == 1.0:
        algo = 'Newton with trust region for simple bound constraints'
    elif proportion_analytical_hessian == 0.0:
        algo = 'BFGS with trust region for simple bound constraints'
    else:
        algo = (
            f'Hybrid Newton [{100*proportion_analytical_hessian}%] with trust '
            f'region for simple bound constraints'
        )

    columns = [Column(title='Iter.', width=5)]

    if variable_names is not None:
        columns += [Column(title=name[:15], width=15) for name in variable_names]
    columns += [
        Column(title='Function', width=12),
        Column(title='Relgrad', width=10),
        Column(title='Radius', width=8),
        Column(title='Rho', width=8),
        Column(title=' ', width=4),
    ]

    the_formatter = FormattedColumns(columns)

    the_hybrid_function = HybridFunction(
        the_function, proportion_analytical_hessian, bounds=bounds
    )

    iterate = bounds.project(starting_point)
    current_function = the_hybrid_function.calculate_function_and_derivatives(iterate)
    if current_function is None:
        messages = the_function.messages
        messages['Algorithm'] = algo
        messages['Number of iterations'] = '0'
        return iterate, messages

    radius = first_radius
    max_delta = np.finfo(float).max
    min_delta = np.finfo(float).eps
    rho = 0.0
    status = ''

    logger.info(the_formatter.formatted_title())

    # Initialize the sample
    the_function.first_sample()
    iterations_with_same_sample = 0

    for k in range(maxiter):

        def logmessage():
            values_to_report = [k]
            if variable_names is not None:
                values_to_report += list(iterate)
            values_to_report += [
                current_function.function,
                the_function.relative_gradient_norm,
                float(radius),
                rho,
                status,
            ]
            logger.info(the_formatter.formatted_row(values_to_report))

        # Solve the quadratic problem in the subspace defined by the GCP
        candidate, _ = bounds.truncated_conjugate_gradient_subspace(
            iterate=iterate,
            gradient=current_function.gradient,
            hessian=current_function.hessian,
            radius=radius,
            tol=cgtol,
        )
        if np.isnan(candidate).any():
            radius = radius / 2.0
            status = '-'
            logmessage()
            continue

        try:
            function_value_candidate = the_hybrid_function.calculate_function(candidate)
            num = current_function.function - function_value_candidate
            step = candidate - iterate
            denom = -np.inner(step, current_function.gradient) - 0.5 * np.inner(
                step, current_function.hessian @ step
            )
            rho = num / denom
            failed = rho < eta1
            if failed:
                radius = min(radius / 2.0, np.linalg.norm(step, np.inf) / 2.0)
                status = '-'
                logmessage()
                continue
        except RuntimeError as e:
            logger.warning(e)
            radius = radius / 2.0
            failed = True
            status = '-'
            logmessage()
            continue

        # Candidate accepted
        try:
            candidate_function = the_hybrid_function.calculate_function_and_derivatives(
                candidate
            )
        except OptimizationError:
            # Failure: reduce the trust region
            radius = min(radius / 2.0, np.linalg.norm(step, np.inf) / 2.0)
            status = '-'
            logmessage()
            continue

        if candidate_function is None:
            # the_matrix stopping criterion has been detected
            messages = the_function.messages
            messages['Algorithm'] = algo
            messages['Number of iterations'] = f'{k+1}'
            messages['Proportion of Hessian calculation'] = (
                the_hybrid_function.message()
            )
            logmessage()
            return candidate, messages

        iterate = candidate
        current_function = candidate_function

        if rho >= eta2:
            # Enlarge the trust region
            radius = min(enlarging_factor * radius, max_delta)
            status = '++'
        else:
            status = '+'

        iterations_with_same_sample += 1
        if iterations_with_same_sample >= 5:
            the_function.change_relative_batch_size(1.10)
            iterations_with_same_sample = 0

        if radius <= min_delta:
            messages = the_function.messages
            messages['Algorithm'] = algo
            message = f'Trust region is too small: {radius}'
            messages['Cause of termination'] = message
            messages['Number of iterations'] = f'{k+1}'
            messages['Proportion of Hessian calculation'] = (
                the_hybrid_function.message()
            )
            logmessage()
            return iterate, messages
        logmessage()

    the_function.calculate_relative_projected_gradient(bounds)
    messages = the_function.messages
    messages['Algorithm'] = algo
    message = f'Maximum number of iterations reached: {maxiter}'
    messages['Cause of termination'] = message
    messages['Number of iterations'] = f'{maxiter}'
    messages['Proportion of Hessian calculation'] = the_hybrid_function.message()
    return iterate, messages



"""File test_algebra.py

:author: Michel Bierlaire
:date: Thu Jun 22 12:36:34 2023

Tests for the algebra module
"""

import unittest

from biogeme_optimization.algebra import (
    schnabel_eskow,
    schnabel_eskow_direction,
)
from biogeme_optimization.exceptions import (
    OptimizationError,
)  # Import the required exception class


class SchnabelEskowTestCase(unittest.TestCase):
    def test_schnabelEskow(self):
        A = np.array([[2, -1, 0], [-1, 2, -1], [0, -1, 2]])
        L, E, P = schnabel_eskow(A)

        # Verify that L is lower triangular
        self.assertTrue(np.all(np.tril(L) == L))

        # Verify that the_matrix + E = PLL^TP^T
        result = P @ L @ L.T @ P.T + E
        np.testing.assert_array_almost_equal(A, result, decimal=5)

        # Test case 1: Identity matrix
        A = np.eye(3)
        L, E, P = schnabel_eskow(A)
        # Verify that L is lower triangular
        self.assertTrue(np.all(np.tril(L) == L))

        expected_L = np.eye(3)
        expected_E = np.zeros(3)
        expected_P = np.eye(3)
        self.assertTrue(np.allclose(L, expected_L))
        self.assertTrue(np.allclose(E, expected_E))
        self.assertTrue(np.allclose(P, expected_P))

        # Verify that the_matrix + E = PLL^TP^T
        result = P @ L @ L.T @ P.T + E
        np.testing.assert_array_almost_equal(A, result, decimal=5)

        # Test case 2: Positive definite matrix
        A = np.array([[4, 2, 1], [2, 5, 2], [1, 2, 6]])
        L, E, P = schnabel_eskow(A)
        # Verify that L is lower triangular
        self.assertTrue(np.all(np.tril(L) == L))

        # Verify that the_matrix + E = PLL^TP^T
        result = P @ L @ L.T @ P.T + E
        np.testing.assert_array_almost_equal(A, result, decimal=5)

        # Test case 3: Non-positive definite matrix
        A = np.array([[1, 1, 1], [1, 2, 2], [1, 2, 3]])
        L, E, P = schnabel_eskow(A)
        # Verify that L is lower triangular
        self.assertTrue(np.all(np.tril(L) == L))

        # Verify that the_matrix + E = PLL^TP^T
        result = P @ L @ L.T @ P.T + E
        np.testing.assert_array_almost_equal(A, result, decimal=5)

        # Test case 4: Larger matrix
        A = np.array([[2, -1, 0, 0], [-1, 2, -1, 0], [0, -1, 2, -1], [0, 0, -1, 2]])
        L, E, P = schnabel_eskow(A)
        # Verify that L is lower triangular
        self.assertTrue(np.all(np.tril(L) == L))

        # Verify that the_matrix + E = PLL^TP^T
        result = P @ L @ L.T @ P.T + E
        np.testing.assert_array_almost_equal(A, result, decimal=5)

        A = np.array(
            [
                [0.3571, -0.1030, 0.0274, -0.0459],
                [-0.1030, 0.2525, 0.0736, -0.3845],
                [0.0274, 0.0736, 0.2340, -0.2878],
                [-0.0459, -0.3845, -0.2878, 0.5549],
            ]
        )
        L, E, P = schnabel_eskow(A)
        self.assertAlmostEqual(L[0][0], 0.7449161, 5)
        diff = P @ L @ L.T @ P.T - E - A
        for i in list(diff.flatten()):
            self.assertAlmostEqual(i, 0.0, 5)

        A = np.array(
            [
                [1890.3, -1705.6, -315.8, 3000.3],
                [-1705.6, 1538.3, 284.9, -2706.6],
                [-315.8, 284.9, 52.5, -501.2],
                [3000.3, -2706.6, -501.2, 4760.8],
            ]
        )
        L, E, P = schnabel_eskow(A)
        self.assertAlmostEqual(L[0][0], 6.89985507e01, 5)
        diff = P @ L @ L.T @ P.T - E - A
        for i in list(diff.flatten()):
            self.assertAlmostEqual(i, 0.0, 5)

    def test_schnabel_eskow_exceptions(self):
        # Test case 1: Non-square matrix
        A = np.array([[1, 2, 3], [4, 5, 6]])
        with self.assertRaises(OptimizationError):
            schnabel_eskow(A)

        # Test case 2: Non-symmetric matrix
        A = np.array([[1, 2, 3], [2, 4, 5], [13, 5, 6]])
        with self.assertRaises(OptimizationError):
            schnabel_eskow(A)


import unittest
import numpy as np


class SchnabelEskowDirectionTestCase(unittest.TestCase):
    def test_schnabel_eskow(self):
        # Define input matrix the_matrix
        A = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

        # Call the schnabel_eskow function
        L, E, P = schnabel_eskow(A)
        # Define the expected results
        expected_L = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
        expected_E = np.zeros_like(expected_L)
        expected_P = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

        # Compare the results with the expected values
        self.assertTrue(np.allclose(L, expected_L))
        self.assertTrue(np.allclose(E, expected_E))
        self.assertTrue(np.allclose(P, expected_P))

    def test_schnabel_eskow_complex(self):
        expected_L = np.array(
            [
                [8.71779789, 0.0, 0.0],
                [0.45883147, 0.37674965, 0.0],
                [2.63828094, 2.09548617, 0.02963117],
            ]
        )
        expected_E = np.array(
            [[0.0, 0.0, 0.0], [0.0, 1.35246662, 0.0], [0.0, 0.0, 1.35246662]]
        )
        expected_P = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

        A = expected_P @ expected_L @ expected_L.T @ expected_P.T - expected_E
        L, E, P = schnabel_eskow(A)
        # Compare the results with the expected values
        self.assertTrue(np.allclose(L, expected_L))
        self.assertTrue(np.allclose(E, expected_E))
        self.assertTrue(np.allclose(P, expected_P))

    def test_schnabel_eskow_non_convex(self):
        # Define input matrix the_matrix
        A = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 1]])

        # Call the schnabel_eskow function
        L, E, P = schnabel_eskow(A)

        # Define the expected results
        positive_values_exist = np.any(E > 0)
        self.assertTrue(positive_values_exist)

    def test_schnabel_eskow_direction(self):
        gradient = np.array([-2.0, 3.0, -1.0])
        hessian = np.array([[4.0, -1.0, 2.0], [-1.0, 6.0, -4.0], [2.0, -4.0, 7.0]])

        result = schnabel_eskow_direction(gradient, hessian)
        # Expected descent direction computed manually
        expected_direction = np.array([0.52808989, -0.6741573, -0.39325843])

        # Perform assertions to validate the output
        np.testing.assert_allclose(result, expected_direction, rtol=1e-6)


if __name__ == '__main__':
    unittest.main()



"""File vns.py

:author: Michel Bierlaire, EPFL
:date: Wed Jul  5 11:50:56 2023

Multi-objective variable neighborhood search algorithm
"""

import logging
import random
from collections import defaultdict

from biogeme_optimization.exceptions import OptimizationError
from biogeme_optimization.neighborhood import Neighborhood
from biogeme_optimization.pareto import Pareto, SetElement

logger: logging.Logger = logging.getLogger(__name__)


class ParetoClass(Pareto):
    """Class managing the solutions"""

    def __init__(self, max_neighborhood: int, pareto_file: str | None = None) -> None:
        """

        :param max_neighborhood: the maximum size of the neighborhood
            that must be considered
        :type max_neighborhood: int

        :param pareto_file: name of a  file containing sets  from previous runs
        :type pareto_file: str

        """
        super().__init__(pareto_file)
        self.max_neighborhood: int = max_neighborhood
        """the maximum size of the neighborhood that must be considered
        """
        self.neighborhood_size: defaultdict[SetElement, int] = defaultdict(int)
        """ dict associating the solutions IDs with the neighborhoodsize"""

        for elem in self.considered:
            self.neighborhood_size[elem] = 1

    def change_neighborhood(self, element: SetElement) -> None:
        """Change the size of the neighborhood for a solution in the Pareto set.

        :param element: ID of the solution for which the neighborhood
            size must be increased.
        :type element: SetElement

        """
        self.neighborhood_size[element] += 1

    def reset_neighborhood(self, element: SetElement) -> None:
        """Reset the size of the neighborhood to 1 for a solution.

        :param element: ID of the solution for which the neighborhood
            size must be reset.
        :type element: biogeme.pareto.SetElement

        """
        self.neighborhood_size[element] = 1

    def add(self, element: SetElement) -> bool:
        """Add an element
        :param element: element to be considered for inclusion in the Pareto set.
        :type element: SetElement

        :return: True if solution has been included. False otherwise.
        :rtype: bool
        """
        added = super().add(element)
        if added:
            self.neighborhood_size[element] = 1
        else:
            self.neighborhood_size[element] += 1
        return added

    def select(self) -> tuple[SetElement | None, int | None]:
        """
        Select a candidate to be modified during the next iteration.

        :return: a candidate and the neighborhood size
        :rtype: tuple(SetElement, int)

        """

        # Candidates are members of the Pareto set that have not
        # reached the maximum neighborhood size.

        candidates = [
            (k, v)
            for k, v in self.neighborhood_size.items()
            if v < self.max_neighborhood
        ]

        if not candidates:
            return None, None
        element, size = random.choice(candidates)
        return element, size


def vns(
    problem: Neighborhood,
    first_solutions: list[SetElement],
    pareto: ParetoClass,
    number_of_neighbors: int = 10,
    maximum_attempts: int = 100,
):
    """Multi objective Variable Neighborhood Search

    :param problem: problem description
    :type problem: neighborhood.NeighborHood

    :param first_solutions: several models to initialize  the algorithm
    :type first_solutions: list(biogeme.pareto.SetElement)

    :param pareto: object managing the Pareto set
    :type pareto: ParetoClass

    :param number_of_neighbors: if none of this neighbors improves the
                              solution, it is considered that a local
                              optimum has been reached.
    :type number_of_neighbors: int

    :param maximum_attempts: an attempts consists in selecting a
        solution in the Pareto set, and trying to improve it. The
        parameters imposes an upper bound on the total number of
        attempts, irrespectively if they are successful or not.

    :type maximum_attempts: int

    :return: the pareto set, the set of models that have been in the
             pareto set and then removed, and the set of all models
             considered by the algorithm.
    :rtype: class :class:`biogeme.vns.ParetoClass`

    :raise OptimizationError: if the first Pareto set is empty.

    """
    print('*** VNS ***')
    if first_solutions is not None:
        for solution in first_solutions:
            valid, why = problem.is_valid(solution)
            if valid:
                pareto.add(solution)
                logger.info(solution)
            else:
                pareto.add_invalid(solution)
                logger.warning(solution)
                logger.warning(f'Default specification is invalid: {why}')

    if pareto.length() == 0:
        raise OptimizationError('Cannot start the algorithm with an empty Pareto set.')

    logger.info(f'Initial pareto: {pareto.length()}')

    solution_to_improve, neighborhood_size = pareto.select()

    pareto.dump()
    for attempt in range(maximum_attempts):
        if solution_to_improve is None:
            break

        logger.info(f'Attempt {attempt}/{maximum_attempts}')
        for neighbor_count in range(number_of_neighbors):
            logger.debug(f'----> Current solution: {solution_to_improve}')
            logger.debug(f'----> Neighbor {neighbor_count} of size {neighborhood_size}')

            a_neighbor, number_of_changes = problem.generate_neighbor(
                solution_to_improve, neighborhood_size
            )
            logger.debug(
                f'----> Neighbor: {a_neighbor} Nbr of changes {number_of_changes}'
            )

            logger.info(
                f'Considering neighbor {neighbor_count}/{number_of_neighbors}'
                f' for current solution'
            )

            if neighbor_count == number_of_neighbors:
                # We have reached the maximum number of neighvors to consider.
                pareto.change_neighborhood(solution_to_improve)
                break

            if number_of_changes == 0:
                # It was not possible to find a neighbor
                logger.info(
                    f'The solution could not be improved with neighborhood '
                    f'of size {neighborhood_size}.'
                )
                pareto.change_neighborhood(solution_to_improve)
                break

            if a_neighbor in pareto.considered:
                # The neighbor has already been considered earlier
                problem.last_neighbor_rejected()
                logger.debug(
                    f'*** Neighbor of size {neighborhood_size}:'
                    f' already considered***'
                )
                pareto.change_neighborhood(solution_to_improve)
                break

            valid, why = problem.is_valid(a_neighbor)
            if not valid:
                # The neighbor is not valid for the problem.
                pareto.add_invalid(a_neighbor)
                logger.debug(
                    f'*** Neighbor of size {neighborhood_size}' f' invalid: {why}***'
                )
                problem.last_neighbor_rejected()
                pareto.change_neighborhood(solution_to_improve)
                break

            if pareto.add(a_neighbor):
                # the_matrix new non dominated solution has been found.
                logger.info('*** New pareto solution: ')
                logger.info(a_neighbor)
                problem.last_neighbor_accepted()
                pareto.reset_neighborhood(solution_to_improve)
                pareto.dump()
                break

            # If we reach this point, it means that the negihbor
            # solution has not been accepted, bacause it is dominated
            # by a previous solution .
            logger.debug(f'*** Neighbor of size ' f'{neighborhood_size}: dominated***')
            problem.last_neighbor_rejected()
            pareto.change_neighborhood(solution_to_improve)
            pareto.dump()

        solution_to_improve, neighborhood_size = pareto.select()

    pareto.dump()
    return pareto



"""File format.py

:author: Michel Bierlaire
:date: Mon Jul  3 12:30:10 2023

Object that formats the information of each iteration
"""

from typing import NamedTuple

import numpy as np


class Column(NamedTuple):
    title: str
    width: int


class FormattedColumns:
    """
    Class for formatting the output of iterations
    """

    def __init__(self, columns: list[Column]) -> None:
        """
        Initialize a FormattedColumns object.

        :param columns: List of Column  representing the columns.
        :type columns: list[Column]
        :raises ValueError: If the length of a column title exceeds the specified width.
        """

        for column in columns:
            if len(column.title) > column.width:
                raise ValueError(
                    f'The length of the title '
                    f'["{column.title}" is {len(column.title)} chars.] '
                    f'exceeds the column width: {column.width}'
                )

        self.columns = columns

    def formatted_title(self) -> str:
        """
        Generate the formatted title.

        :return: The formatted title.
        :rtype: str
        """
        return ' '.join(col.title.rjust(col.width) for col in self.columns)

    def formatted_row(self, values: list[str | float | int]) -> str:
        """
        Generate a formatted row based on the given values.

        :param values: List of values representing the row data.
        :type values: list[str, int, float]

        :return: The formatted row.
        :rtype: str

        :raises ValueError: If the number of values doesn't match the number of columns.
        """
        if len(values) != len(self.columns):
            raise ValueError(
                f"There are {len(values)} values for {len(self.columns)} columns"
            )

        return ' '.join(
            self._format_value(col, value) for col, value in zip(self.columns, values)
        )

    @staticmethod
    def _format_value(column: Column, value: str | float | int) -> str:
        """
        Format a canonical_value based on the column type.

        :param column: The column for the canonical_value.
        :type column: Column

        :param value: The canonical_value to be formatted.
        :type value: str, int, float

        :return: The formatted canonical_value.
        :rtype: str

        :raises ValueError: If the canonical_value has an unsupported type.
        """
        if isinstance(value, str):
            return value.rjust(column.width)

        if isinstance(value, float):
            return f'{value:{column.width}.2g}'

        if isinstance(value, int) or isinstance(value, np.integer):
            return f'{value:{column.width}d}'

        raise ValueError(
            f'Invalid column type: {type(value)} for canonical_value {value}'
        )



"""File test_trust_region.py

:author: Michel Bierlaire
:date: Fri Jun 23 09:32:05 2023

Tests for the trust_region module
"""

import unittest
import numpy as np
import scipy.linalg as la
from biogeme_optimization.exceptions import (
    OptimizationError,
)  # Import the required exception class
from biogeme_optimization.trust_region import (
    trust_region_intersection,
    cauchy_newton_dogleg,
    dogleg,
    truncated_conjugate_gradient,
    newton_trust_region,
    bfgs_trust_region,
    DoglegDiagnostic,
    ConjugateGradientDiagnostic,
    QuadraticModel,
    NewtonModel,
    BfgsModel,
)
from examples import MyFunctionToMinimize, Example58, Rosenbrock


class TrustRegionIntersectionTest(unittest.TestCase):
    def test_intersection_1(self):
        """See example p. 295 in Bierlaire (2015)"""
        xhat = np.array([9, 1])
        xc = np.array([9, 1])
        dc = xc - xhat
        xd = np.array([7.2, -0.8])
        dd = xd - xhat
        radius = 1
        inside_direction = xc - xhat
        crossing_direction = xd - xc

        expected_intersection = np.array([8.29289, 0.29289])
        expected_lambda = la.norm(expected_intersection - xc) / la.norm(xd - xc)
        result = trust_region_intersection(inside_direction, crossing_direction, radius)

        self.assertAlmostEqual(result, expected_lambda, places=3)

    def test_intersection_2(self):
        """See example p. 295 in Bierlaire (2015)"""
        xhat = np.array([9, 1])
        xc = np.array([7.2, -0.8])
        dc = xc - xhat
        xd = np.array([4.608, 0.512])
        dd = xd - xhat
        radius = 4
        inside_direction = xc - xhat
        crossing_direction = xd - xc

        expected_intersection = np.array([5.06523, 0.28056])
        expected_lambda = la.norm(expected_intersection - xc) / la.norm(xd - xc)
        result = trust_region_intersection(inside_direction, crossing_direction, radius)

        self.assertAlmostEqual(result, expected_lambda, places=3)

    def test_intersection_3(self):
        """See example p. 295 in Bierlaire (2015)"""
        xhat = np.array([9, 1])
        xc = np.array([9, 1])
        dc = xc - xhat
        xd = np.array([0, 0])
        dd = xd - xhat
        radius = 8
        inside_direction = xc - xhat
        crossing_direction = xd - xc

        expected_intersection = np.array([1.04893, 0.11655])
        expected_lambda = la.norm(expected_intersection - xc) / la.norm(xd - xc)
        result = trust_region_intersection(inside_direction, crossing_direction, radius)

        self.assertAlmostEqual(result, expected_lambda, places=3)

    def no_test_intersection(self):
        """See example p. 295 in Bierlaire (2015)"""
        xhat = np.array([9, 1])
        xc = np.array([9, 1])
        dc = xc - xhat
        xd = np.array([0, 0])
        dd = xd - xhat
        radius = 18
        inside_direction = xc - xhat
        crossing_direction = xd - xc

        with self.assertRaises(OptimizationError):
            _ = trust_region_intersection(inside_direction, crossing_direction, radius)

    def test_intersection_xc_outside(self):
        xhat = np.array([9, 1])
        xc = np.array([7.2, -0.8])
        dc = xc - xhat
        xd = np.array([4.608, 0.512])
        dd = xd - xhat
        radius = 0.1
        inside_direction = xc - xhat
        crossing_direction = xd - xc

        with self.assertRaises(OptimizationError):
            result = trust_region_intersection(
                inside_direction, crossing_direction, radius
            )

    def test_intersection_xd_inside(self):
        xhat = np.array([9, 1])
        xc = np.array([7.2, -0.8])
        dc = xc - xhat
        xd = np.array([4.608, 0.512])
        dd = xd - xhat
        radius = 18
        inside_direction = xc - xhat
        crossing_direction = xd - xc

        with self.assertRaises(OptimizationError):
            result = trust_region_intersection(
                inside_direction, crossing_direction, radius
            )

    def test_cauchy_newton_dogleg(self):
        g = np.array([2, -1])  # Example gradient
        H = np.array([[4, -1], [-1, 2]])  # Example Hessian

        # Expected results
        d_cauchy_expected = np.array([-0.45454545, 0.22727273])
        d_newton_expected = np.array([-0.42857143, 0.28571429])
        d_dogleg_expected = np.array([-0.42662338, 0.28441558])

        # Calculate actual results
        d_cauchy, d_newton, d_dogleg = cauchy_newton_dogleg(g, H)
        normalized_cauchy = d_cauchy / np.linalg.norm(d_cauchy)
        normalized_gradient = -g / np.linalg.norm(g)
        np.testing.assert_allclose(normalized_gradient, normalized_cauchy)
        normalized_newton = d_newton / np.linalg.norm(d_newton)
        normalized_dogleg = d_dogleg / np.linalg.norm(d_dogleg)
        np.testing.assert_allclose(normalized_gradient, normalized_cauchy)
        # Compare actual and expected results
        np.testing.assert_allclose(d_cauchy, d_cauchy_expected)
        np.testing.assert_allclose(d_newton, d_newton_expected)
        np.testing.assert_allclose(d_dogleg, d_dogleg_expected)

    def test_cauchy_newton_dogleg_simple(self):
        g = np.array([1, 1])  # Example gradient
        H = np.array([[1, 0], [0, 1]])  # Example Hessian

        # Expected results
        d_cauchy_expected = np.array([-1, -1])
        d_newton_expected = np.array([-1, -1])
        d_dogleg_expected = np.array([-1, -1])

        # Calculate actual results
        d_cauchy, d_newton, d_dogleg = cauchy_newton_dogleg(g, H)
        normalized_cauchy = d_cauchy / np.linalg.norm(d_cauchy)
        normalized_gradient = -g / np.linalg.norm(g)
        np.testing.assert_allclose(normalized_gradient, normalized_cauchy)
        normalized_newton = d_newton / np.linalg.norm(d_newton)
        normalized_dogleg = d_dogleg / np.linalg.norm(d_dogleg)
        np.testing.assert_allclose(normalized_gradient, normalized_cauchy)
        # Compare actual and expected results
        np.testing.assert_allclose(d_cauchy, d_cauchy_expected)
        np.testing.assert_allclose(d_newton, d_newton_expected)
        np.testing.assert_allclose(d_dogleg, d_dogleg_expected)

    def test_cauchy_newton_dogleg_non_convex(self):
        g = np.array([1, 1])  # Example gradient
        H = np.array([[1, 0], [0, -1]])  # Example Hessian
        d_cauchy, d_newton, d_dogleg = cauchy_newton_dogleg(g, H)
        self.assertIsNone(d_newton)
        self.assertIsNone(d_dogleg)
        self.assertIsNone(d_cauchy)

    def test_cauchy_newton_dogleg_non_convex_2(self):
        g = np.array([1, 0])  # Example gradient
        H = np.array([[1, 0], [0, -1]])  # Example Hessian
        d_cauchy, d_newton, d_dogleg = cauchy_newton_dogleg(g, H)
        self.assertIsNone(d_newton)
        self.assertIsNone(d_dogleg)
        d_cauchy_expected = np.array([-1, 0])
        np.testing.assert_allclose(d_cauchy, d_cauchy_expected)

    def test_dogleg(self):
        # Test case 1: Negative curvature along Cauchy direction
        gradient = np.array([1, 0])
        hessian = np.array([[-1, 0], [0, -1]])
        radius = 1.0
        expected_solution = np.array([-1, 0])
        expected_diagnostic = DoglegDiagnostic.NEGATIVE_CURVATURE
        solution, diagnostic = dogleg(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

        # Test case 2: Negative curvature along Newton direction
        gradient = np.array([2, -1])
        hessian = np.array([[4, -1], [-1, -3]])
        radius = 3
        expected_solution = np.array([-0.58823529, 0.29411765])
        expected_diagnostic = DoglegDiagnostic.CAUCHY
        solution, diagnostic = dogleg(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

        # Test case 3: Partial Cauchy step
        xhat = np.array([9, 1])
        gradient = np.array([9, 9])
        hessian = np.array([[1, 0], [0, 9]])
        radius = 1
        xc = np.array([7.2, -0.8])
        xsol = np.array([8.29289, 0.29289])
        expected_solution = xsol - xhat
        expected_diagnostic = DoglegDiagnostic.PARTIAL_CAUCHY
        solution, diagnostic = dogleg(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(solution, expected_solution, decimal=3)
        self.assertEqual(diagnostic, expected_diagnostic)

        # Test case 4: Dogleg
        xhat = np.array([9, 1])
        gradient = np.array([9, 9])
        hessian = np.array([[1, 0], [0, 9]])
        radius = 4
        xc = np.array([7.2, -0.8])
        xsol = np.array([5.06523, 0.28056])
        expected_solution = xsol - xhat
        expected_diagnostic = DoglegDiagnostic.DOGLEG
        solution, diagnostic = dogleg(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(solution, expected_solution, decimal=3)
        self.assertEqual(diagnostic, expected_diagnostic)

        # Test case 5: partial Newton
        xhat = np.array([9, 1])
        gradient = np.array([9, 9])
        hessian = np.array([[1, 0], [0, 9]])
        radius = 8
        xc = np.array([7.2, -0.8])
        xsol = np.array([1.04893, 0.11655])
        expected_solution = xsol - xhat
        expected_diagnostic = DoglegDiagnostic.PARTIAL_NEWTON
        solution, diagnostic = dogleg(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(solution, expected_solution, decimal=3)
        self.assertEqual(diagnostic, expected_diagnostic)

        # Test case 10: Newton
        xhat = np.array([9, 1])
        gradient = np.array([9, 9])
        hessian = np.array([[1, 0], [0, 9]])
        radius = 10
        xc = np.array([7.2, -0.8])
        xsol = np.array([0, 0])
        expected_solution = xsol - xhat
        expected_diagnostic = DoglegDiagnostic.NEWTON
        solution, diagnostic = dogleg(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(solution, expected_solution, decimal=3)
        self.assertEqual(diagnostic, expected_diagnostic)

    def test_dogleg_invalid_input(self):
        # Test invalid input: gradient and hessian are not NumPy arrays
        with self.assertRaises(OptimizationError):
            dogleg([1, 2, 3], [[1, 0, 0], [0, 1, 0], [0, 0, 1]], 1.0)

        # Test invalid input: gradient is not 1-dimensional or hessian is not 2-dimensional
        with self.assertRaises(OptimizationError):
            dogleg(np.array([1, 2, 3]), np.array([[1, 0], [0, 1], [0, 0]]), 1.0)

        # Test invalid input: gradient and hessian dimensions are incompatible
        with self.assertRaises(OptimizationError):
            dogleg(np.array([1, 2, 3]), np.array([[1, 0, 0], [0, 1, 0]]), 1.0)

    def test_conjugate_gradient_convergence(self):
        gradient = np.array([1, 2, 3])
        hessian = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        radius = 10.0
        expected_step = np.array([-1, -2, -3])
        expected_diagnostic = ConjugateGradientDiagnostic.CONVERGENCE

        step, diagnostic = truncated_conjugate_gradient(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(step, expected_step)
        self.assertEqual(diagnostic, expected_diagnostic)

    def test_conjugate_gradient_out_of_trust_region(self):
        gradient = np.array([1, 2, 3])
        hessian = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        radius = 1.0
        expected_step = np.array([-0.26726124, -0.53452248, -0.80178373])
        expected_diagnostic = ConjugateGradientDiagnostic.OUT_OF_TRUST_REGION

        step, diagnostic = truncated_conjugate_gradient(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(step, expected_step)
        self.assertEqual(diagnostic, expected_diagnostic)

    def test_conjugate_gradient_negative_curvature(self):
        gradient = np.array([1, 2, 3])
        hessian = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 1]])
        radius = 10.0
        expected_step = np.array([-2.56752763, -5.83763813, -7.70258288])
        expected_diagnostic = ConjugateGradientDiagnostic.NEGATIVE_CURVATURE

        step, diagnostic = truncated_conjugate_gradient(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(step, expected_step)
        self.assertEqual(diagnostic, expected_diagnostic)

    def test_conjugate_gradient_numerical_problem(self):
        gradient = np.array([1, 2, 3])
        hessian = np.array([[1, 0, 0], [0, 10e12, 0], [0, 0, -10e12]])
        radius = 10.0
        expected_step = np.array([-2.67261242, -5.34522484, -8.01783726])
        expected_diagnostic = ConjugateGradientDiagnostic.NEGATIVE_CURVATURE
        step, diagnostic = truncated_conjugate_gradient(gradient, hessian, radius)
        np.testing.assert_array_almost_equal(step, expected_step)
        self.assertEqual(diagnostic, expected_diagnostic)


class TestQuadraticModel(unittest.TestCase):
    def test_f_g_h(self):
        function = MyFunctionToMinimize(dimension=2)
        with self.assertRaises(TypeError):
            _ = QuadraticModel(function)


class TestNewtonModel(unittest.TestCase):
    def test_f_g_h(self):
        function = MyFunctionToMinimize(dimension=2)
        model = NewtonModel(function)
        iterate = np.array([1, 1])
        f, g, h = model.get_f_g_h(iterate)
        expected_g = np.array([1, 1])
        np.testing.assert_array_almost_equal(g, expected_g)

        expected_h = np.array([[1.0, 0.0], [0.0, 1.0]])
        np.testing.assert_array_almost_equal(h, expected_h)

        optimal_solution = np.array([0, 0])
        f, g, h = model.get_f_g_h(optimal_solution)
        self.assertIsNone(g)
        self.assertIsNone(h)


class TestBfgsModel(unittest.TestCase):
    def test_f_g_h(self):
        function = MyFunctionToMinimize(dimension=2)
        model = BfgsModel(function)
        iterate = np.array([1, 1])
        f, g, h = model.get_f_g_h(iterate)

        expected_g = np.array([1, 1])
        np.testing.assert_array_almost_equal(g, expected_g)

        expected_h = np.array([[1.0, 0.0], [0.0, 1.0]])
        np.testing.assert_array_almost_equal(h, expected_h)

        optimal_solution = np.array([0, 0])
        f, g, h = model.get_f_g_h(optimal_solution)
        self.assertIsNone(g)
        self.assertIsNone(h)


class TestAlgorithms(unittest.TestCase):
    def test_newton_trust_region(self):
        fct = MyFunctionToMinimize(dimension=2)

        # Set the starting point
        starting_point = np.array([2.0, 2.0])

        # Set the expected solution
        expected_solution = np.array([0.0, 0.0])

        # Call the newton_linesearch function
        optimization_results = newton_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=True
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 0 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

        # Call the newton_linesearch function
        optimization_results = newton_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=False
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 0 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

    def test_bfgs_trust_region(self):
        fct = MyFunctionToMinimize(dimension=2)

        # Set the starting point
        starting_point = np.array([2.0, 2.0])

        # Set the expected solution
        expected_solution = np.array([0.0, 0.0])

        # Call the newton_linesearch function
        optimization_results = bfgs_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=True
        )
        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )
        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 0 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

        # Call the newton_linesearch function
        optimization_results = bfgs_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=False
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 0 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

    def test_newton_trust_region_ex58(self):
        fct = Example58()

        # Set the starting point
        starting_point = np.array([1.0, 1.0])

        # Set the expected solution
        expected_solution = np.array([-1.0, 0])

        # Call the newton_linesearch function
        optimization_results = newton_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=True
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-3
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 7.1e-07 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

        # Call the newton_linesearch function
        optimization_results = newton_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=False
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 1.5e-09 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

    def test_bfgs_trust_region_ex58(self):
        fct = Example58()

        # Set the starting point
        starting_point = np.array([1.0, 1.0])

        # Set the expected solution
        expected_solution = np.array([-1, 0])

        # Call the newton_linesearch function
        optimization_results = bfgs_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=True
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-3
        )

        # Check the termination message
        termination_mesage = optimization_results.messages['Cause of termination']
        expected_termination_message = 'Relative gradient = ...'
        self.assertEqual(termination_mesage[:19], expected_termination_message[:19])
        self.assertTrue(optimization_results.convergence)

        # Call the newton_linesearch function
        optimization_results = bfgs_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=False
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-3
        )

        # Check the termination message
        termination_mesage = optimization_results.messages['Cause of termination']
        expected_termination_message = 'Relative gradient = ...'
        self.assertEqual(termination_mesage[:19], expected_termination_message[:19])
        self.assertTrue(optimization_results.convergence)

    def test_newton_trust_region_rosenbrock(self):
        fct = Rosenbrock()

        # Set the starting point
        starting_point = np.array([-1.5, 1.5])

        # Set the expected solution
        expected_solution = np.array([1.0, 1.0])

        # Call the newton_linesearch function
        optimization_results = newton_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=True
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-3
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 2.6e-06 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

        # Call the newton_linesearch function
        optimization_results = newton_trust_region(
            the_function=fct, starting_point=starting_point, use_dogleg=False
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-6
        )

        # Check the termination message
        self.assertEqual(
            optimization_results.messages['Cause of termination'],
            'Relative gradient = 9.8e-10 <= 6.1e-06',
        )
        self.assertTrue(optimization_results.convergence)

    def test_bfgs_trust_region_rosenbrock(self):
        fct = Rosenbrock()

        # Set the starting point
        starting_point = np.array([-1.5, 1.5])

        # Set the expected solution
        expected_solution = np.array([1.0, 1.0])

        # Call the newton_linesearch function
        optimization_results = bfgs_trust_region(
            the_function=fct,
            starting_point=starting_point,
            use_dogleg=True,
            maxiter=20000,
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-3
        )

        # Check the termination message
        termination_mesage = optimization_results.messages['Cause of termination']
        expected_termination_message = 'Relative gradient = ...'
        self.assertEqual(termination_mesage[:19], expected_termination_message[:19])
        self.assertTrue(optimization_results.convergence)

        # Call the newton_linesearch function
        optimization_results = bfgs_trust_region(
            the_function=fct,
            starting_point=starting_point,
            use_dogleg=False,
            maxiter=20000,
        )

        # Check the solution
        np.testing.assert_allclose(
            optimization_results.solution, expected_solution, atol=1e-3
        )

        # Check the termination message
        termination_mesage = optimization_results.messages['Cause of termination']
        expected_termination_message = 'Relative gradient = ...'
        self.assertEqual(termination_mesage[:19], expected_termination_message[:19])
        self.assertTrue(optimization_results.convergence)


if __name__ == '__main__':
    unittest.main()



"""File trust_region.py

:author: Michel Bierlaire
:date: Fri Jun 23 09:27:17 2023

Functions for trust region algorithms
"""

import logging
from abc import ABC, abstractmethod
from typing import Callable

import numpy as np
from biogeme_optimization.diagnostics import (
    OptimizationResults,
    DoglegDiagnostic,
    ConjugateGradientDiagnostic,
    Diagnostic,
)
from biogeme_optimization.exceptions import OptimizationError
from biogeme_optimization.algebra import schnabel_eskow_direction
from biogeme_optimization.bfgs import bfgs
from biogeme_optimization.function import FunctionToMinimize, FunctionData

logger = logging.getLogger(__name__)


class QuadraticModel(ABC):
    """Abstract class for the generation of the quadratic model."""

    def __init__(self, function: FunctionToMinimize):
        """Constructor

        :param function: the function to minimize
        :type function: FunctionToMinimize

        """
        self.the_function: FunctionToMinimize = function

    @abstractmethod
    def get_f_g_h(self, iterate: np.ndarray) -> FunctionData:
        """Obtain the vector g and the matrix h characterizing the
            model, as well as the canonical_value of the function

        :param iterate: current iterate
        :type iterate: numpy.array (n x 1)

        :return: function, gradient and hessian
        :rtype: tuple(float, numpy.array (nx1), numpy.array (nxn)
        """


class NewtonModel(QuadraticModel):
    """Call implementing the quadratic model based on the analytical hessian"""

    def get_f_g_h(self, iterate: np.ndarray) -> FunctionData:
        """Obtain the vector g and the matrix h characterizing the model

        :param iterate: current iterate
        :type iterate: numpy.array (n x 1)

        :return: function, gradient and hessian
        :rtype: tuple(float, numpy.array (nx1), numpy.array (nxn)
        """
        self.the_function.set_variables(iterate)
        evaluation = self.the_function.f_g_h()
        optimal = self.the_function.check_optimality()
        if optimal:
            return FunctionData(function=None, gradient=None, hessian=None)
        return evaluation


class BfgsModel(QuadraticModel):
    """Call implementing the quadratic model based on the BFGS approximation"""

    def __init__(
        self,
        function: FunctionToMinimize,
        first_approximation: np.ndarray | None = None,
    ):
        """Constructor

        :param first_approximation: first approximation of the Hessian
        :type first_approximation: numpy.array (n x n)

        :return: function, gradient and hessian
        :rtype: tuple(float, numpy.array (nx1), numpy.array (nxn)
        """
        super().__init__(function)

        if first_approximation is None:
            self.hessian_approx: np.ndarray = np.identity(self.the_function.dimension())
        else:
            if first_approximation.ndim != self.the_function.dimension():
                error_msg = (
                    f'Incompatible dimensions: {self.the_function.dimension()} '
                    f'and {first_approximation.ndim}'
                )
                raise OptimizationError(error_msg)
            self.hessian_approx = first_approximation
        self.last_iterate: np.ndarray | None = None
        self.last_gradient: np.ndarray | None = None

    def get_f_g_h(self, iterate: np.ndarray) -> FunctionData:
        """Obtain the vector g and the matrix h characterizing the model

        :param iterate: current iterate
        :type iterate: numpy.array (n x 1)

        """
        self.the_function.set_variables(iterate)
        evaluation = self.the_function.f_g()
        optimal = self.the_function.check_optimality()
        if optimal:
            return FunctionData(function=None, gradient=None, hessian=None)
        # The first time, there is nothing to update
        if self.last_iterate is not None:
            delta_x = iterate - self.last_iterate
            delta_g = evaluation.gradient - self.last_gradient
            # Update the approximation, if possible
            try:
                self.hessian_approx = bfgs(self.hessian_approx, delta_x, delta_g)
            except OptimizationError as err:
                logger.warning(err)
        self.last_iterate = iterate
        self.last_gradient = evaluation.gradient
        return FunctionData(
            function=evaluation.function,
            gradient=evaluation.gradient,
            hessian=self.hessian_approx,
        )


def trust_region_intersection(
    inside_direction: np.ndarray,
    crossing_direction: np.ndarray,
    radius: float,
    check_step: bool = True,
) -> float:
    """Calculates the intersection with the boundary of the trust region.

    Consider a trust region of radius :math:`\\delta`, centered at
    :math:`\\hat{x}`. Let :math:`x_c` be in the trust region, and
    :math:`d_c = x_c - \\hat{x}`, so that :math:`\\|d_c\\| \\leq
    \\delta`. Let :math:`x_d` be out of the trust region, and
    :math:`d_d = x_d - \\hat{x}`, so that :math:`\\|d_d\\| \\geq
    \\delta`.  We calculate :math:`\\lambda` such that

    .. math:: \\| d_c + \\lambda (d_d - d_c)\\| = \\delta

    :param inside_direction: xc - xhat.
    :type inside_direction: numpy.array

    :param crossing_direction: dd - dc
    :type crossing_direction: numpy.array

    :param radius: radius of the trust region.
    :type radius: float

    :param check_step: if True, an exception is raised if the canonical_value of
        the step is out of the interval [0,1]
    :type check_step: bool

    :return: :math:`\\lambda` such that :math:`\\| d_c +
              \\lambda (d_d - d_c)\\| = \\delta`

    :rtype: float

    """
    a = np.inner(crossing_direction, crossing_direction)
    b = 2 * np.inner(inside_direction, crossing_direction)
    c = np.inner(inside_direction, inside_direction) - radius**2
    discriminant = b * b - 4.0 * a * c
    if discriminant < 0:
        raise OptimizationError('No intersection')
    step = (-b + np.sqrt(discriminant)) / (2 * a)
    if check_step:
        if step < 0:
            raise OptimizationError('Starting point outside trust region')
        if step > 1:
            raise OptimizationError('Target point inside trust region')
    return step


def cauchy_newton_dogleg(
    gradient: np.ndarray, hessian: np.ndarray
) -> tuple[np.ndarray | None, np.ndarray | None, np.ndarray | None]:
    """Calculate the Cauchy, the Newton and the dogleg points.

    The Cauchy point is defined as

    .. math:: d_c = - \\frac{\\nabla f(x)^T \\nabla f(x)}
                    {\\nabla f(x)^T \\nabla^2 f(x)
                    \\nabla f(x)} \\nabla f(x)

    The Newton point :math:`d_n` verifies Newton equation:

    .. math:: H_s d_n = - \\nabla f(x)

    where :math:`H_s` is a positive definite matrix generated with the
    method by `Schnabel and Eskow (1999)`_.

    The Dogleg point is

    .. math:: d_d = \\eta d_n

    where

    .. math:: \\eta = 0.2 + 0.8 \\frac{\\alpha^2}{\\beta |\\nabla f(x)^T d_n|}

    and :math:`\\alpha= \\nabla f(x)^T \\nabla f(x)`,
    :math:`\\beta=\\nabla f(x)^T \\nabla^2 f(x)\\nabla f(x).`

    :param gradient: gradient :math:`\\nabla f(x)`

    :type gradient: numpy.array

    :param hessian: hessian :math:`\\nabla^2 f(x)`

    :type hessian: numpy.array

    :return: tuple with Cauchy point, Newton point, Dogleg point. The
        Cauchy point in None if the curvature is negative along the
        gradient direction. The Newton and the Dogleg directions are
        None if the curvature is negative along the Newton direction.
    :rtype: numpy.array, numpy.array, numpy.array

    """
    alpha = np.inner(gradient, gradient)
    beta = np.inner(gradient, hessian @ gradient)
    if beta > 0:
        d_cauchy = -(alpha / beta) * gradient
    else:
        d_cauchy = None
    try:
        d_newton = schnabel_eskow_direction(
            gradient=gradient, hessian=hessian, check_convexity=True
        )
    except OptimizationError:
        return d_cauchy, None, None
    eta = 0.2 + (0.8 * alpha * alpha / (beta * abs(np.inner(gradient, d_newton))))
    return d_cauchy, d_newton, eta * d_newton


def dogleg(
    gradient: np.ndarray, hessian: np.ndarray, radius: float
) -> tuple[np.ndarray, DoglegDiagnostic]:
    """
    Find an approximation of the trust region sub-problem using
    the dogleg method

    :param gradient: gradient of the quadratic model.
    :type gradient: numpy.array
    :param hessian: hessian of the quadratic model.
    :type hessian: numpy.array
    :param radius: radius of the trust region.
    :type radius: float

    :return: d, diagnostic where

          - d is an approximate solution of the trust region subproblem
          - diagnostic is the nature of the solution:

             * NEGATIVE_CURVATURE if negative curvature along Cauchy direction
               (i.e. along the gradient)
             * CAUCHY if Cauchy step
             * PARTIAL_CAUCHY partial Cauchy step
             * NEWTON if Newton step
             * PARTIAL_NEWTON if partial Newton step
             * DOGLEG if Dogleg

    :rtype: numpy.array, int
    """

    # Check if the inputs are valid NumPy arrays
    if not isinstance(gradient, np.ndarray) or not isinstance(hessian, np.ndarray):
        raise OptimizationError('gradient and hessian must be NumPy arrays')
    if gradient.ndim != 1 or hessian.ndim != 2:
        raise OptimizationError(
            'gradient must be a 1-dimensional array and '
            'hessian must be a 2-dimensional array'
        )
    if gradient.shape[0] != hessian.shape[0] or hessian.shape[0] != hessian.shape[1]:
        raise OptimizationError("gradient and hessian must have compatible dimensions")

    cauchy_dir, newton_dir, dogleg_dir = cauchy_newton_dogleg(gradient, hessian)

    # Check if the model is convex along the gradient direction
    if cauchy_dir is None:
        d_star = -radius * gradient / np.linalg.norm(gradient)
        return d_star, DoglegDiagnostic.NEGATIVE_CURVATURE

    alpha = np.inner(gradient, gradient)
    beta = np.inner(gradient, hessian @ gradient)
    norm_dc = alpha * np.sqrt(alpha) / beta
    if norm_dc >= radius:
        # The Cauchy point is outside the trust
        # region. We move along the Cauchy
        # direction until the border of the trust
        # region.
        d_star = (radius / norm_dc) * cauchy_dir
        return d_star, DoglegDiagnostic.PARTIAL_CAUCHY

    # Check the convexity of the model along Newton direction
    if newton_dir is None:
        # Return the Cauchy point
        return cauchy_dir, DoglegDiagnostic.CAUCHY

    # Compute Newton point
    norm_dn = np.linalg.norm(newton_dir)

    if norm_dn <= radius:
        # Newton point is inside the trust region
        return newton_dir, DoglegDiagnostic.NEWTON

    # Compute the dogleg point

    eta = 0.2 + (0.8 * alpha * alpha / (beta * abs(np.inner(gradient, newton_dir))))

    partial_newton = eta * norm_dn

    if partial_newton <= radius:
        # Dogleg point is inside the trust region
        d_star = (radius / norm_dn) * newton_dir
        return d_star, DoglegDiagnostic.PARTIAL_NEWTON

    # Between Cauchy and dogleg
    crossing_direction = dogleg_dir - cauchy_dir
    lbd = trust_region_intersection(cauchy_dir, crossing_direction, radius)
    d_star = cauchy_dir + lbd * crossing_direction
    return d_star, DoglegDiagnostic.DOGLEG


def truncated_conjugate_gradient(
    gradient: np.ndarray, hessian: np.ndarray, radius: float, tol: float = 1.0e-6
) -> tuple[np.ndarray, ConjugateGradientDiagnostic]:
    """Find an approximation of the trust region sub-problem using the
    truncated conjugate gradient method

    :param gradient: gradient of the quadratic model.
    :type gradient: numpy.array

    :param hessian: hessian of the quadratic model.
    :type hessian: numpy.array

    :param radius: radius of the trust region.
    :type radius: float

    :param tol: tolerance on the norm of the gradient, used as a stopping criterion
    :type tol: float

    :return: d, diagnostic, where

          - d is the approximate solution of the trust region sub-problem,
          - diagnostic is the nature of the solution:

            * CONVERGENCE for convergence,
            * OUT_OF_TRUST_REGION if out of the trust region,
            * NEGATIVE_CURVATURE if negative curvature detected.
            * NUMERICAL_PROBLEM if a numerical problem has been encountered

    :rtype: numpy.array, ConjugateGradientDiagnostic

    """
    dimension = len(gradient)
    xk = np.zeros(dimension)
    gk = gradient
    dk = -gk
    for _ in range(dimension):
        try:
            curvature = np.inner(dk, hessian @ dk)
            if curvature <= 0:
                # Negative curvature has been detected
                diagnostic = ConjugateGradientDiagnostic.NEGATIVE_CURVATURE
                step = trust_region_intersection(xk, dk, radius, check_step=False)
                solution = xk + step * dk
                return solution, diagnostic
            alphak = -np.inner(dk, gk) / curvature
            xkp1 = xk + alphak * dk
            if np.isnan(xkp1).any() or np.linalg.norm(xkp1) > radius:
                # Out of the trust region
                diagnostic = ConjugateGradientDiagnostic.OUT_OF_TRUST_REGION
                step = trust_region_intersection(xk, dk, radius, check_step=False)
                solution = xk + step * dk
                return solution, diagnostic
            xk = xkp1
            gkp1 = hessian @ xk + gradient
            betak = np.inner(gkp1, gkp1) / np.inner(gk, gk)
            dk = -gkp1 + betak * dk
            gk = gkp1
            if np.linalg.norm(gkp1) <= tol:
                diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
                step = xk
                return step, diagnostic
        except ValueError:
            # Numerical problem. We follow the last direction until
            # the border of the trust region
            diagnostic = ConjugateGradientDiagnostic.NUMERICAL_PROBLEM
            step = trust_region_intersection(xk, dk, radius, check_step=False)
            solution = xk + step * dk
            return solution, diagnostic
    diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
    step = xk
    return step, diagnostic


def minimization_with_trust_region(
    the_function: FunctionToMinimize,
    starting_point: np.ndarray,
    quadratic_model: QuadraticModel,
    solving_trust_region_subproblem: Callable[
        [np.ndarray, np.ndarray, float], tuple[np.ndarray, Diagnostic]
    ],
    maxiter: int = 1000,
    initial_radius: float = 1.0,
    eta1: float = 0.01,
    eta2: float = 0.9,
):
    """Minimization method with trust region

    :param the_function: object to calculate the objective function and its derivatives.
    :type the_function: optimization.functionToMinimize

    :param starting_point: starting point
    :type starting_point: numpy.array

    :param quadratic_model: object in charge of generating the quadratic model
    :type quadratic_model: QuadraticModel

    :param solving_trust_region_subproblem: function in charge of
        solving the trust region sub-problem


    :param maxiter: the algorithm stops if this number of iterations
                    is reached. Default: 1000.
    :type maxiter: int

    :param initial_radius: initial radius of the trust region.
    :type initial_radius: float

    :param eta1: threshold for failed iterations. Default: 0.01.
    :type eta1: float

    :param eta2: threshold for very successful iterations. Default 0.9.
    :type eta2: float

    :return: named tuple:

            - solution is the solution found,
            - message is a dictionary reporting various aspects
              related to the run of the algorithm.
            - convergence is a bool which is True if the algorithm has converged

    :rtype: OptimizationResults

    """
    xk = starting_point
    the_output: FunctionData = quadratic_model.get_f_g_h(xk)
    value_iterate = the_output.function
    g = the_output.gradient
    h = the_output.hessian
    radius = initial_radius
    max_delta = np.finfo(float).max
    min_delta = np.finfo(float).eps
    for k in range(maxiter):
        # If the iterate is optimal, the returned canonical_value is None
        if value_iterate is None:
            messages = the_function.messages
            messages['Number of iterations'] = k
            optimization_results = OptimizationResults(
                solution=xk, messages=messages, convergence=True
            )
            break
        step, _ = solving_trust_region_subproblem(g, h, radius)
        candidate = xk + step
        the_function.set_variables(candidate)
        # Calculate the canonical_value of the function
        value_candidate = the_function.f()
        if value_candidate >= value_iterate:
            radius = np.linalg.norm(step) / 2.0
            status = '-'
            continue

        num = value_iterate - value_candidate
        denominator = -np.inner(step, g) - 0.5 * np.inner(step, h @ step)
        rho = num / denominator
        if rho < eta1:
            # Failure: reduce the trust region
            radius = np.linalg.norm(step) / 2.0
            status = '-'
            continue

        # Candidate accepted
        xk = candidate
        the_output: FunctionData = quadratic_model.get_f_g_h(xk)
        value_iterate = the_output.function
        g = the_output.gradient
        h = the_output.hessian
        if rho >= eta2:
            # Enlarge the trust region
            radius = min(2 * radius, max_delta)
            status = '++'
        else:
            status = '+'
        if radius <= min_delta:
            messages = the_function.messages
            messages['Cause of termination'] = f'Trust region is too small: {radius}'
            optimization_results = OptimizationResults(
                solution=xk, messages=messages, convergence=False
            )
            break
    else:
        messages = the_function.messages
        messages['Cause of termination'] = (
            f'Maximum number of iterations reached: {maxiter}'
        )
        optimization_results = OptimizationResults(
            solution=xk, messages=messages, convergence=False
        )

    return optimization_results


def newton_trust_region(
    *,
    the_function: FunctionToMinimize,
    starting_point: np.ndarray,
    use_dogleg: bool = False,
    maxiter: int = 1000,
    initial_radius: float = 1.0,
    eta1: float = 0.01,
    eta2: float = 0.9,
):
    """Newton method with trust region

    :param the_function: object to calculate the objective function and its derivatives.
    :type the_function: optimization.functionToMinimize

    :param starting_point: starting point
    :type starting_point: numpy.array

    :param use_dogleg: True if the trust region sub-problem is solved using
        the dogleg method. False is it is solved with the truncated
        conjugate gradient algorithm.
    :type use_dogleg: bool

    :param maxiter: the algorithm stops if this number of iterations
                    is reached. Default: 1000.
    :type maxiter: int

    :param initial_radius: initial radius of the trust region. Default: 100.
    :type initial_radius: float

    :param eta1: threshold for failed iterations. Default: 0.01.
    :type eta1: float

    :param eta2: threshold for very successful iterations. Default 0.9.
    :type eta2: float

    :return: named tuple:

            - solution is the solution found,
            - message is a dictionary reporting various aspects
              related to the run of the algorithm.
            - convergence is a bool which is True if the algorithm has converged

    :rtype: OptimizationResults

    """
    the_model = NewtonModel(the_function)

    if use_dogleg:
        return minimization_with_trust_region(
            the_function,
            starting_point,
            the_model,
            dogleg,
            maxiter,
            initial_radius,
            eta1,
            eta2,
        )
    return minimization_with_trust_region(
        the_function,
        starting_point,
        the_model,
        truncated_conjugate_gradient,
        maxiter,
        initial_radius,
        eta1,
        eta2,
    )


def bfgs_trust_region(
    *,
    the_function: FunctionToMinimize,
    starting_point: np.ndarray,
    init_bfgs: np.ndarray | None = None,
    use_dogleg: bool = False,
    maxiter: int = 1000,
    initial_radius: float = 1.0,
    eta1: float = 0.01,
    eta2: float = 0.9,
):
    """BFGS method with trust region

    :param the_function: object to calculate the objective function and its derivatives.
    :type the_function: optimization.functionToMinimize

    :param starting_point: starting point
    :type starting_point: numpy.array

    :param init_bfgs: matrix used to initialize BFGS. If None, the
                     identity matrix is used. Default: None.
    :type init_bfgs: numpy.array

    :param use_dogleg: True if the trust region subproblem is solved using
        the dogleg method. False is it is solved with the truncated
        conjugate gradient algorithm.
    :type use_dogleg: bool

    :param maxiter: the algorithm stops if this number of iterations
                    is reached. Default: 1000.
    :type maxiter: int

    :param initial_radius: initial radius of the trust region. Default: 100.
    :type initial_radius: float

    :param eta1: threshold for failed iterations. Default: 0.01.
    :type eta1: float

    :param eta2: threshold for very successful iterations. Default 0.9.
    :type eta2: float

    :return: named tuple:

            - solution is the solution found,
            - message is a dictionary reporting various aspects
              related to the run of the algorithm.
            - convergence is a bool which is True if the algorithm has converged

    :rtype: OptimizationResults

    """
    the_model = BfgsModel(the_function, init_bfgs)

    if use_dogleg:
        return minimization_with_trust_region(
            the_function,
            starting_point,
            the_model,
            dogleg,
            maxiter,
            initial_radius,
            eta1,
            eta2,
        )
    return minimization_with_trust_region(
        the_function,
        starting_point,
        the_model,
        truncated_conjugate_gradient,
        maxiter,
        initial_radius,
        eta1,
        eta2,
    )



"""File test_vns.py

:author: Michel Bierlaire
:date: Fri Jul  7 15:30:29 2023

Tests for the vns module
"""
import unittest
import os
from collections import defaultdict
import tempfile
from unittest.mock import MagicMock, patch
from random import choice
from knapsack import Knapsack, Sack
from biogeme_optimization.vns import ParetoClass, vns
from biogeme_optimization.pareto import SetElement


class ParetoClassTests(unittest.TestCase):
    def setUp(self):
        self.max_neighborhood = 5
        self.pareto_file = tempfile.mktemp()
        self.pareto_class = ParetoClass(self.max_neighborhood, self.pareto_file)
        self.element1 = SetElement(1, [1, 2])
        self.element2 = SetElement(2, [2, 1])

    def tearDown(self):
        if os.path.exists(self.pareto_file):
            os.remove(self.pareto_file)

    def test_init(self):
        self.assertEqual(self.pareto_class.max_neighborhood, self.max_neighborhood)
        self.assertEqual(self.pareto_class.filename, self.pareto_file)
        self.assertEqual(self.pareto_class.neighborhood_size, defaultdict(int))

    def test_change_neighborhood(self):
        self.pareto_class.change_neighborhood(self.element1)
        self.assertEqual(self.pareto_class.neighborhood_size[self.element1], 1)
        self.pareto_class.change_neighborhood(self.element1)
        self.assertEqual(self.pareto_class.neighborhood_size[self.element1], 2)

    def test_reset_neighborhood(self):
        self.pareto_class.change_neighborhood(self.element1)
        self.pareto_class.reset_neighborhood(self.element1)
        self.assertEqual(self.pareto_class.neighborhood_size[self.element1], 1)

    def test_add_element_added(self):
        self.pareto_class.add(self.element1)
        self.assertEqual(self.pareto_class.neighborhood_size[self.element1], 1)

    def test_add_element_not_added(self):
        self.pareto_class.add(self.element1)
        self.pareto_class.add(self.element1)
        self.assertEqual(self.pareto_class.neighborhood_size[self.element1], 2)

    def test_select(self):
        candidate = (self.element1, 3)
        self.pareto_class.neighborhood_size = {self.element1: 1, self.element2: 3}
        with patch("random.choice", MagicMock(return_value=candidate)):
            selected_element, neighborhood_size = self.pareto_class.select()
            self.assertEqual(selected_element, self.element1)
            self.assertEqual(neighborhood_size, 3)

    def test_select_no_candidates(self):
        self.pareto_class.neighborhood_size = {self.element1: 5, self.element2: 5}
        with patch("random.choice", MagicMock(side_effect=IndexError)):
            selected_element, neighborhood_size = self.pareto_class.select()
            self.assertIsNone(selected_element)
            self.assertIsNone(neighborhood_size)


class VNSTests(unittest.TestCase):
    def setUp(self):
        self.UTILITY = [80, 31, 48, 17, 27, 84, 34, 39, 46, 58, 23, 67]
        self.WEIGHT = [84, 27, 47, 22, 21, 96, 42, 46, 54, 53, 32, 78]
        self.CAPACITY = 300
        self.FILE_NAME = 'knapsack.pareto'
        Sack.utility_data = self.UTILITY
        Sack.weight_data = self.WEIGHT
        self.the_pareto = ParetoClass(max_neighborhood=5, pareto_file=self.FILE_NAME)
        self.the_knapsack = Knapsack(self.UTILITY, self.WEIGHT, self.CAPACITY)
        self.empty_sack = Sack.empty_sack(size=len(self.UTILITY))

    def test_vns(self):
        the_pareto = vns(
            problem=self.the_knapsack,
            first_solutions=[self.empty_sack.get_element()],
            pareto=self.the_pareto,
            number_of_neighbors=5,
        )


if __name__ == "__main__":
    unittest.main()
